"""MPF clock and main loop."""
import asyncio
from typing import Tuple
from serial_asyncio import create_serial_connection
from mpf.core.logging import LogMixin


class PeriodicTask:

    """A periodic asyncio task."""

    __slots__ = ["_canceled", "_interval", "_callback", "_loop", "_last_call"]

    def __init__(self, interval, loop, callback):
        """Initialise periodic task."""
        self._canceled = False
        self._interval = interval
        self._callback = callback
        self._loop = loop
        self._last_call = self._loop.time()
        self._schedule()

    def _schedule(self):
        if self._canceled:
            return
        self._loop.call_at(self._last_call + self._interval, self._run)

    def get_next_call_time(self):
        """Return time of next call."""
        return self._last_call + self._interval

    def _run(self):
        self._last_call = self._last_call + self._interval
        if self._canceled:
            return
        self._callback()
        self._schedule()

    def cancel(self):
        """Cancel periodic task."""
        self._canceled = True


class ClockBase(LogMixin):

    """A clock object with event support."""

    __slots__ = ["machine", "loop"]

    def __init__(self, machine=None, loop=None):
        """Initialise clock."""
        super().__init__()
        self.machine = machine

        # needed since the test clock is setup before the machine
        if machine:
            self.configure_logging(
                'Clock',
                self.machine.config['logging']['console']['clock'],
                self.machine.config['logging']['file']['clock'])
        else:
            self.configure_logging('Clock', None, None)

        self.debug_log("Starting tickless clock")
        if not loop:
            self.loop = self._create_event_loop()   # type: asyncio.AbstractEventLoop
        else:
            self.loop = loop                        # type: asyncio.AbstractEventLoop

        asyncio.set_event_loop(self.loop)

    # pylint: disable-msg=no-self-use
    def _create_event_loop(self):
        try:
            # pylint: disable-msg=import-outside-toplevel
            import uvloop
        except ImportError:
            pass
        else:
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        return asyncio.get_event_loop()

    def run(self, stop_future):
        """Run the clock."""
        return self.loop.run_until_complete(stop_future)

    def get_time(self):
        """Get the last tick made by the clock."""
        return self.loop.time()

    def start_server(self, client_connected_cb, host=None, port=None, **kwd):
        """Start a server."""
        return asyncio.start_server(client_connected_cb, host, port, loop=self.loop, **kwd)

    def open_connection(self, host=None, port=None, *,
                        limit=None, **kwds):
        """Open connection using asyncio.

        Wrapper for create_connection() returning a (reader, writer) pair.

        The reader returned is a StreamReader instance; the writer is a
        StreamWriter instance.

        The arguments are all the usual arguments to create_connection()
        except protocol_factory; most common are positional host and port,
        with various optional keyword arguments following.

        Additional optional keyword arguments are loop (to set the event loop
        instance to use) and limit (to set the buffer limit passed to the
        StreamReader).

        (If you want to customize the StreamReader and/or
        StreamReaderProtocol classes, just copy the code -- there's
        really nothing special here except some convenience.)
        """
        if not limit:
            # pylint: disable-msg=protected-access
            limit = asyncio.streams._DEFAULT_LIMIT
        return asyncio.open_connection(host=host, port=port, loop=self.loop, limit=limit, **kwds)

    async def open_serial_connection(self, limit=None, **kwargs) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Open a serial connection using asyncio.

        A wrapper for create_serial_connection() returning a (reader, writer) pair.

        The reader returned is a StreamReader instance; the writer is a StreamWriter instance.

        The arguments are all the usual arguments to Serial(). Additional
        optional keyword arguments are loop (to set the event loop instance
        to use) and limit (to set the buffer limit passed to the
        StreamReader.

        This function is a coroutine.

        Args:
        ----
            loop: asyncio loop
            limit: line length limit
        """
        if not limit:
            # pylint: disable-msg=protected-access
            limit = asyncio.streams._DEFAULT_LIMIT      # type: ignore

        reader = asyncio.StreamReader(limit=limit, loop=self.loop)
        protocol = asyncio.StreamReaderProtocol(reader, loop=self.loop)
        transport, _ = await create_serial_connection(
            loop=self.loop,
            protocol_factory=lambda: protocol,
            **kwargs)
        writer = asyncio.StreamWriter(transport, protocol, reader, self.loop)
        return reader, writer

    def schedule_once(self, callback, timeout=0):
        """Schedule an event in <timeout> seconds.

        If <timeout> is unspecified
        or 0, the callback will be called after the next frame is rendered.

        Args:
        ----
            callback: callback to call on timeout
            timeout: seconds to wait

        Returns a :class:`ClockEvent` instance.
        """
        if not callable(callback):
            raise AssertionError('callback must be a callable, got %s' % callback)

        event = self.loop.call_later(delay=timeout, callback=callback)

        if self._debug_to_console or self._debug_to_file:
            self.debug_log("Scheduled a one-time clock callback (callback=%s, timeout=%s)",
                           str(callback), timeout)

        return event

    def schedule_interval(self, callback, timeout) -> PeriodicTask:
        """Schedule an event to be called every <timeout> seconds.

        Args:
        ----
            callback: callback to call on timeout
            timeout: period to wait

        Returns a PeriodicTask object.
        """
        if not callable(callback):
            raise AssertionError('callback must be a callable, got {}'.format(callback))

        periodic_task = PeriodicTask(timeout, self.loop, callback)

        if self._debug_to_console or self._debug_to_file:
            self.debug_log("Scheduled a recurring clock callback (callback=%s, timeout=%s)",
                           str(callback), timeout)

        return periodic_task

    @staticmethod
    def unschedule(event):
        """Remove a previously scheduled event. Wrapper for cancel for compatibility to kivy clock.

        Args:
        ----
            event: Event to cancel
        """
        try:
            event.cancel()
        except Exception:     # pylint: disable-msg=broad-except
            raise AssertionError("Broken unschedule: {} {}".format(event, type(event)))



from mpf.core.config_validator import ValidationPath

from mpf.core.utility_functions import Util
from mpf.exceptions.config_file_error import ConfigFileError
from mpf.tests.MpfTestCase import MpfTestCase
from mpf._version import log_url


class TestConfig(MpfTestCase):

    def get_config_file(self):
        return 'test_config_interface.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/config_interface/'

    def setUp(self):
        self.machine_spec_patches['test_section'] = dict(__valid_in__='machine')
        super().setUp()

    def test_config_file(self):
        # true, True, yes, Yes values should be True
        self.assertIs(True, self.machine.config['test_section']['true_key1'])
        self.assertIs(True, self.machine.config['test_section']['true_key2'])
        self.assertIs(True, self.machine.config['test_section']['true_key3'])
        self.assertIs(True, self.machine.config['test_section']['true_key4'])

        # false, False, no, No values should be False
        self.assertIs(False, self.machine.config['test_section']['false_key1'])
        self.assertIs(False, self.machine.config['test_section']['false_key2'])
        self.assertIs(False, self.machine.config['test_section']['false_key3'])
        self.assertIs(False, self.machine.config['test_section']['false_key4'])

        # on, off values should be strings
        self.assertEqual('on', self.machine.config['test_section']['on_string'])
        self.assertEqual('off', self.machine.config['test_section']['off_string'])

        # 6400, 6, 07 should be ints
        self.assertEqual(6400, self.machine.config['test_section']['int_6400'])
        self.assertEqual(6, self.machine.config['test_section']['int_6'])
        self.assertEqual(7, self.machine.config['test_section']['int_7'])

        # 00ff00, 003200 should be stringsc
        self.assertEqual('00ff00', self.machine.config['test_section']['str_00ff00'])
        self.assertEqual('003200', self.machine.config['test_section']['str_003200'])

        # +5, +0.5 should be strings
        self.assertEqual('+5', self.machine.config['test_section']['str_plus5'])
        self.assertEqual('+0.5', self.machine.config['test_section']['str_plus0point5'])

        # keys should keep case
        self.assertIn('case_sensitive_1', self.machine.config['test_section'])
        self.assertIn('Case_sensitive_2', self.machine.config['test_section'])
        self.assertIn('case_sensitive_3', self.machine.config['test_section'])

        # values should be case sensitive
        self.assertEqual(self.machine.config['test_section']['case_sensitive_1'], 'test')
        self.assertEqual(self.machine.config['test_section']['Case_sensitive_2'], 'test')
        self.assertEqual(self.machine.config['test_section']['case_sensitive_3'], 'Test')

    def test_config_validator(self):
        validation_failure_info = ValidationPath(ValidationPath(ValidationPath(None, "key"), "entry"), "subkey")
        # test config spec syntax error
        self.assertRaises(ValueError,
                          self.machine.config_validator.validate_config_item,
                          'single|int'.split("|"), None, None)

        # test default required, source is int
        validation_string = 'single|int|'.split("|")
        results = self.machine.config_validator.validate_config_item(
                validation_string, 'test_failure_info', 0)
        self.assertEqual(results, 0)

        # test default provided, source overrides default
        validation_string = 'single|int|0'.split("|")
        results = self.machine.config_validator.validate_config_item(
                validation_string, 'test_failure_info', 1)
        self.assertEqual(results, 1)

        # test source type is converted to int
        validation_string = 'single|int|0'.split("|")
        results = self.machine.config_validator.validate_config_item(
                validation_string, 'test_failure_info', '1')
        self.assertEqual(results, 1)

        # test default when no source is present
        validation_string = 'single|int|1'.split("|")
        results = self.machine.config_validator.validate_config_item(
                validation_string, 'test_failure_info')  # no item in config
        self.assertEqual(results, 1)

        # test default required with source missing raises error
        validation_string = 'single|int|'.split("|")  # default required
        self.assertRaises(ConfigFileError,
                          self.machine.config_validator.validate_config_item,
                          validation_string, ValidationPath(None, "test"))  # no item

        # test broken int
        validation_string = 'single|int|'.split("|")  # default required
        with self.assertRaises(AssertionError) as e:
            results = self.machine.config_validator.validate_config_item(
                validation_string, validation_failure_info, '1s')

        # test int with range
        validation_string = 'single|int(0,10)|'.split("|")

        results = self.machine.config_validator.validate_config_item(
            validation_string, validation_failure_info, '0')
        self.assertEqual(results, 0)

        results = self.machine.config_validator.validate_config_item(
            validation_string, validation_failure_info, '7')
        self.assertEqual(results, 7)

        results = self.machine.config_validator.validate_config_item(
            validation_string, validation_failure_info, '10')
        self.assertEqual(results, 10)

        with self.assertRaises(AssertionError) as e:
            results = self.machine.config_validator.validate_config_item(
                validation_string, validation_failure_info, '-1')

        with self.assertRaises(AssertionError) as e:
            results = self.machine.config_validator.validate_config_item(
                validation_string, validation_failure_info, 11)

        # int with open end
        validation_string = 'single|int(0,NONE)|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, validation_failure_info, 999)
        self.assertEqual(results, 999)

        with self.assertRaises(AssertionError) as e:
            results = self.machine.config_validator.validate_config_item(
                validation_string, validation_failure_info, -1)

        # test str validations

        validation_string = 'single|str|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'hello')
        self.assertEqual(results, 'hello')
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 1)
        self.assertEqual(results, '1')
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', "none")
        self.assertEqual(results, None)

        # test lstr
        validation_string = 'single|lstr|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'HellO')
        self.assertEqual(results, 'hello')
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'hello')
        self.assertEqual(results, 'hello')
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'H1')
        self.assertEqual(results, 'h1')
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'none')
        self.assertEqual(results, None)

        # test float validations

        validation_string = 'single|float|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 1)
        self.assertAlmostEqual(results, 1.0, .01)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '1')
        self.assertAlmostEqual(results, 1.0, .01)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 1.0)
        self.assertAlmostEqual(results, 1.0, .01)

        # test num validations

        validation_string = 'single|num|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 1.0)
        self.assertAlmostEqual(results, 1.0, .01)
        self.assertEqual(type(results), float)
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '1.0')
        self.assertAlmostEqual(results, 1.0, .01)
        self.assertEqual(type(results), float)
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 1)
        self.assertEqual(results, 1)
        self.assertIs(type(results), int)
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '1')
        self.assertEqual(results, 1)
        self.assertIs(type(results), int)

        # test int_from_hex
        validation_string = 'single|int_from_hex|ff'.split("|")
        results = self.machine.config_validator.validate_config_item(
                validation_string, 'test_failure_info')  # no item in config
        self.assertEqual(results, 255)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '0a')
        self.assertEqual(results, 10)

        with self.assertRaises(AssertionError):
            self.machine.config_validator.validate_config_item(
                validation_string, validation_failure_info, 'white')

        # test bool validations
        validation_string = 'single|bool|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'f')
        self.assertFalse(results)

        validation_string = 'single|boolean|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'f')
        self.assertFalse(results)

        validation_string = 'single|bool|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'false')
        self.assertFalse(results)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', False)
        self.assertFalse(results)

        # test bool_int validations
        validation_string = 'single|bool_int|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'f')
        self.assertEqual(0, results)
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 't')
        self.assertEqual(1, results)
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'false')
        self.assertEqual(0, results)
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', False)
        self.assertEqual(0, results)
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', "True")
        self.assertEqual(1, results)

        # test ms conversions
        validation_string = 'single|ms|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 100)
        self.assertEqual(results, 100)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 100.0)
        self.assertEqual(results, 100)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '100')
        self.assertEqual(results, 100)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '100ms')
        self.assertEqual(results, 100)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '1s')
        self.assertEqual(results, 1000)

        # test sec conversions
        validation_string = 'single|secs|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 100)
        self.assertEqual(results, 100)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 100.0)
        self.assertEqual(results, 100)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '100')
        self.assertEqual(results, 100)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '100s')
        self.assertEqual(results, 100)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '100ms')
        self.assertEqual(results, .1)

        # test single list conversions
        # (this just test it gets converted to a list since string_to_list
        # is tested earlier)
        validation_string = 'single|list|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'hi')
        self.assertEqual(results, ['hi'])

        # Test lists
        validation_string = 'list|int|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '1, 2, 3')
        self.assertEqual(results, [1, 2, 3])

        # Test set
        validation_string = 'set|int|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '1, 2, 3')
        self.assertEqual(results, {1, 2, 3})

        # Test dict
        validation_string = 'dict|str:int|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', dict(hello='1'))
        self.assertEqual(results, dict(hello=1))

        # Test color
        validation_string = 'single|color|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'red')
        self.assertEqual(results, (255, 0, 0))

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'ff0000')
        self.assertEqual(results, (255, 0, 0))

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '255, 0, 0')
        self.assertEqual(results, (255, 0, 0))

        # Test kivycolor
        validation_string = 'single|kivycolor|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'red')
        self.assertEqual(results, [1, 0, 0, 1])

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'ff0000')
        self.assertEqual(results, [1, 0, 0, 1])

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '255, 0, 0')
        self.assertEqual(results, [1, 0, 0, 1])

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 'ff000000')
        self.assertEqual(results, [1, 0, 0, 0])

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '255, 0, 0, 255')
        self.assertEqual(results, [1, 0, 0, 1])

        # Test gain
        validation_string = 'single|gain|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '0.0')
        self.assertEqual(results, 0.0)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '1.0')
        self.assertEqual(results, 1.0)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '2.0')
        self.assertEqual(results, 1.0)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '-3')
        self.assertEqual(results, 0.0)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '0 db')
        self.assertEqual(results, 1.0)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '-inf')
        self.assertEqual(results, 0.0)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '-3 DB')
        self.assertAlmostEqual(results, 0.707945784)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '-6db')
        self.assertAlmostEqual(results, 0.501187233)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '-17.5db')
        self.assertAlmostEqual(results, 0.133352143)

        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '3db')
        self.assertEqual(results, 1.0)

        # test pow2
        validation_string = 'single|pow2|'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', 128)
        self.assertEqual(results, 128)
        results = self.machine.config_validator.validate_config_item(
            validation_string, 'test_failure_info', '128')
        self.assertEqual(results, '128')

        with self.assertRaises(AssertionError):
            self.machine.config_validator.validate_config_item(
                validation_string, validation_failure_info, '127')

        # test enum
        validation_string = 'single|enum(None,test)|None'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, validation_failure_info, None)
        self.assertEqual(None, results)

        results = self.machine.config_validator.validate_config_item(
            validation_string, validation_failure_info, "None")
        self.assertEqual(None, results)

        results = self.machine.config_validator.validate_config_item(
            validation_string, validation_failure_info, "test")
        self.assertEqual("test", results)

        self.maxDiff = None
        with self.assertRaises(AssertionError) as e:
            self.machine.config_validator.validate_config_item(
                validation_string, validation_failure_info, 'something else')
        self.assertEqual('Config File Error in ConfigValidator: Config validation error: Entry '
                         'key:entry:subkey = "something else" is not valid. Entry "something else" is not valid for enum. '
                         'Valid values are: None,test Error Code: CFE-ConfigValidator-5 '
                         '({})'.format(log_url.format("CFE-ConfigValidator-5")), str(e.exception))

        # test enum with 'NO' setting
        validation_string = 'single|enum(nc,no)|no'.split("|")
        results = self.machine.config_validator.validate_config_item(
            validation_string, validation_failure_info, False)
        self.assertEqual('no', results)

    def test_config_merge(self):
        a = {"test": {"a": [1], "b": [2, 3]}, "test2": 2}
        b = {"test": {"a": [3], "c": 7}}
        c = Util.dict_merge(a, b)
        self.assertEqual({'test': {'a': [1, 3], 'b': [2, 3], 'c': 7}, 'test2': 2}, c)

        b2 = {"test": {"_overwrite": True, "a": [3], "c": 7}}
        c = Util.dict_merge(a, b2)
        self.assertEqual({'test': {'a': [3], 'c': 7}, 'test2': 2}, c)

        b3 = {"test": {"_delete": True}}
        c = Util.dict_merge(a, b3)
        self.assertEqual({'test2': 2}, c)



"""Test dual wound coil."""
from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings

from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestDualWoundCoil(MpfTestCase):

    def get_config_file(self):
        return 'config_dual_wound_coil.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/device/'

    def testBasicFunctions(self):
        c_power = self.machine.coils["c_power"].hw_driver
        c_hold = self.machine.coils["c_hold"].hw_driver
        c_power.enable = MagicMock()
        c_power.disable = MagicMock()
        c_power.pulse = MagicMock()
        c_hold.enable = MagicMock()
        c_hold.disable = MagicMock()
        c_hold.pulse = MagicMock()

        # test enable
        self.machine.coils["c_test"].enable()
        c_power.pulse.assert_called_with(PulseSettings(power=1.0, duration=20))
        c_power.pulse = MagicMock()
        assert not c_power.enable.called
        c_hold.enable.assert_called_with(PulseSettings(power=1.0, duration=10), HoldSettings(power=1.0))
        c_hold.enable = MagicMock()
        assert not c_hold.pulse.called

        # test disable
        self.machine.coils["c_test"].disable()
        c_power.disable.assert_called_with()
        c_power.disable = MagicMock()
        c_hold.disable.assert_called_with()
        c_hold.disable = MagicMock()

        # test pulse
        self.machine.coils["c_test"].pulse(17)
        c_power.pulse.assert_called_with(PulseSettings(power=1.0, duration=17))
        c_hold.pulse.assert_called_with(PulseSettings(power=1.0, duration=17))

        # test default pulse
        self.machine.coils["c_test"].pulse()
        c_power.pulse.assert_called_with(PulseSettings(power=1.0, duration=20))
        c_hold.pulse.assert_called_with(PulseSettings(power=1.0, duration=10))



from unittest.mock import MagicMock
from mpf.tests.MpfTestCase import MpfTestCase


class TestStepper(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/stepper/'

    def test_linearstepper(self):
        # full range servo
        stepper = self.machine.steppers["linearAxis_stepper"]
        # verify - now at zero both in stepper and in HW
        self.assertEqual(0.0, stepper._current_position)

        # go to min
        commandPos = stepper.config['pos_min']
        stepper._move_to_absolute_position(commandPos)
        self.machine.clock.loop.run_until_complete(
            self.machine.events.wait_for_event("stepper_linearAxis_stepper_ready"))
        self.assertAlmostEqual(commandPos, stepper._current_position, 0)

        # go to max
        commandPos = stepper.config['pos_max']
        stepper._move_to_absolute_position(commandPos)
        self.machine.clock.loop.run_until_complete(
            self.machine.events.wait_for_event("stepper_linearAxis_stepper_ready"))
        self.assertAlmostEqual(commandPos, stepper._current_position, 0)

        # try out of bounds
        with self.assertRaises(ValueError):
            commandPos = stepper.config['pos_min'] - 0.01
            stepper._move_to_absolute_position(commandPos)

        with self.assertRaises(ValueError):
            commandPos = stepper.config['pos_max'] + 0.01
            stepper._move_to_absolute_position(commandPos)

        # relative +/-
        stepper._move_to_absolute_position(0)
        self.machine.clock.loop.run_until_complete(
            self.machine.events.wait_for_event("stepper_linearAxis_stepper_ready"))
        self.assertEqual(0, stepper._current_position)

        # try relative out of bounds
        # at zero, a relative move of min or max limit is equiv to absolute
        stepper._move_to_absolute_position(0)
        self.assertEqual(0, stepper._current_position)

    def test_rotary(self):
        return
        stepper = self.machine.steppers["rotaryMotor_stepper"]

        # spin clockwise
        stepper.move_vel_mode( 1000 )
        self.assertEqual( 1000, stepper._cachedVelocity )

        # stop
        stepper.move_vel_mode( 0 )
        self.assertEqual( 0, stepper._cachedVelocity )

        # spin counter clockwise
        stepper.move_vel_mode( -1000 )
        self.assertEqual( -1000, stepper._cachedVelocity )

        # stop
        stepper.move_vel_mode( 0 )
        self.assertEqual( 0, stepper._cachedVelocity )

        # try to go too fast
        with self.assertRaises(ValueError):
            stepper.move_vel_mode( stepper.config['velocity_limit'] + 0.01 )

        # try a home / should be rejected
        with self.assertRaises(RuntimeError):
            stepper.home()

        # try a position move / should be rejected
        with self.assertRaises(RuntimeError):
            stepper._move_to_absolute_position(42)

        # try a relative move / should be rejected
        with self.assertRaises(RuntimeError):
            stepper.move_rel_pos( 42 )

    def test_stepper_events(self):
        stepper = self.machine.steppers["linearAxis_stepper"]

        # post reset event
        event_future = self.machine.events.wait_for_event("stepper_linearAxis_stepper_ready")
        self.post_event("test_reset")
        self.machine.clock.loop.run_until_complete(event_future)
        # should go to reset position
        self.assertEqual(0.0, stepper._current_position)

        # post another defined event
        event_future = self.machine.events.wait_for_event("stepper_linearAxis_stepper_ready")
        self.post_event("test_00")
        self.machine.clock.loop.run_until_complete(event_future)
        self.assertEqual(-5.0, stepper._current_position, 0)

        # post another defined event
        event_future = self.machine.events.wait_for_event("stepper_linearAxis_stepper_ready")
        self.post_event("test_01")
        self.machine.clock.loop.run_until_complete(event_future)
        self.assertEqual(999.0, stepper._current_position, 0)

        # post another defined event
        event_future = self.machine.events.wait_for_event("stepper_linearAxis_stepper_ready")
        self.post_event("test_10")
        self.machine.clock.loop.run_until_complete(event_future)
        self.assertEqual(500.0, stepper._current_position, 0)

    def test_ball_search(self):
        stepper = self.machine.steppers["linearAxis_stepper"]

        self.machine.playfields["playfield"].config['enable_ball_search'] = True
        self.machine.playfields["playfield"].balls += 1

        event_future = self.machine.events.wait_for_event("stepper_linearAxis_stepper_ready")
        self.post_event("test_10")
        self.machine.clock.loop.run_until_complete(event_future)
        self.assertEqual(500, stepper._current_position)

        # wait until ball search started
        event_future = self.machine.events.wait_for_event("ball_search_started")
        self.machine.clock.loop.run_until_complete(event_future)

        # it will first go to ball search max
        self.advance_time_and_run(.5)
        self.assertEqual(0, stepper._current_position)

        self.advance_time_and_run(5)
        # and then to min
        self.assertEqual(1, stepper._current_position)

        # wait until ball search failed
        event_future = self.machine.events.wait_for_event("ball_search_failed")
        self.machine.clock.loop.run_until_complete(event_future)

        # stepper should restore to previous location
        self.advance_time_and_run(.5)
        self.assertEqual(500, stepper._current_position)



"""Test openpixel hardware interface."""
from mpf.core.rgb_color import RGBColor
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.tests.loop import MockSocket


class TestOpenpixel(MpfTestCase):
    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/openpixel/'

    def get_platform(self):
        return 'openpixel'

    def setUp(self):
        self._messages = []
        super().setUp()
        self.assertOpenPixelLedsSent({}, {})
        self.assertTrue(self._mock_socket.is_open)

    def _mock_loop(self):
        self._mock_socket = MockSocket()
        self.clock.mock_socket("localhost", 7890, self._mock_socket)
        # connect socket to test
        self._mock_socket.send = self._send_mock

    def _build_message(self, channel, leds):
        out = bytearray()
        out.extend([channel, 0, 1, 44])
        for i in range(0, 100):
            if i in leds:
                out.extend([leds[i][0], leds[i][1], leds[i][2]])
            else:
                out.extend([0, 0, 0])

        return bytes(out)

    def _send_mock(self, message):
        self._messages.append(message)
        return len(message)

    def assertOpenPixelLedsSent(self, leds1, leds2):
        if leds1 is not None and leds2 is not None:
            self.assertEqual([self._build_message(0, leds1),
                              self._build_message(1, leds2)],
                             self._messages)
        elif leds1 is not None:
            self.assertEqual([self._build_message(0, leds1)],
                             self._messages)
        elif leds2 is not None:
            self.assertEqual([self._build_message(1, leds2)],
                             self._messages)
        else:
            raise AssertionError("Invalid assert")
        self._messages = []

    def test_led_color(self):
        # test led on channel 0. position 99
        self.machine.lights["test_led"].on()
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent({99: (255, 255, 255)}, None)

        # test led 20 ond channel 0
        self.machine.lights["test_led2"].color(RGBColor((255, 0, 0)))
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent({20: (255, 0, 0), 99: (255, 255, 255)}, None)

        self.machine.lights["test_led"].off()
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent({20: (255, 0, 0), 99: (0, 0, 0)}, None)
        self._messages = []

        # test led color
        self.machine.lights["test_led"].color(RGBColor((2, 23, 42)))
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent({20: (255, 0, 0), 99: (2, 23, 42)}, None)

        # test led on channel 1
        self.machine.lights["test_led3"].on()
        self.advance_time_and_run(1)
        self.assertOpenPixelLedsSent(None, {99: (255, 255, 255)})



# pylint: disable-msg=too-many-lines
"""Contains AssetManager, AssetLoader, and Asset base classes."""
import copy
import os
import random
import time

import threading
from collections import deque, namedtuple
from pathlib import PurePath

from typing import Iterable, Optional, Set, Callable, Tuple, Union
from typing import List

from mpf.core.mode import Mode

from mpf.core.machine import MachineController
from mpf.core.mpf_controller import MpfController
from mpf.core.utility_functions import Util
from mpf.core.logging import LogMixin
from mpf.core.placeholder_manager import BoolTemplate


AssetClass = namedtuple("AssetClass", ["attribute", "cls", "path_string", "config_section", "disk_asset_section",
                                       "extensions", "priority", "pool_config_section", "defaults"])


class BaseAssetManager(MpfController, LogMixin):

    """Base class for the Asset Manager."""

    # needed here so the auto-detection of child classes works
    module_name = 'AssetManager'
    config_name = 'asset_manager'

    __slots__ = ["_asset_classes", "num_assets_to_load", "num_assets_loaded", "num_bcp_assets_to_load",
                 "num_bcp_assets_loaded", "_next_id", "_last_asset_event_time", "initial_assets_loaded", "_start_time"]

    def __init__(self, machine: MachineController) -> None:
        """Initialise asset manager.

        Args:
        ----
            machine: The machine controller
        """
        super().__init__(machine)

        self.machine.register_boot_hold('assets')

        self._asset_classes = list()        # type: List[AssetClass]
        # List of dicts, with each dict being an asset class. See
        # register_asset_class() method for details.

        self.initial_assets_loaded = False  # type: bool
        # True if the initial asset loaded is done

        self.num_assets_to_load = 0
        # Total number of assets that are/will be loaded. Used for
        # calculating progress. Reset to 0 when num_assets_loaded matches it.

        self.num_assets_loaded = 0
        # Number of assets loaded so far. Reset to 0 when it hits
        # num_assets_to_load.

        self.num_bcp_assets_to_load = 0
        self.num_bcp_assets_loaded = 0

        self._next_id = 0
        # id of next asset

        self.machine.mode_controller.register_start_method(
            start_method=self._load_mode_assets)

        # Modes load in init_phase_1, so by 2 we have everything to create
        # the assets.
        if 'force_assets_load' in self.machine.options:
            force_assets_load = self.machine.options['force_assets_load']
        else:
            force_assets_load = False
        self.machine.events.add_handler('init_phase_3', self._create_assets,
                                        force_assets_load=force_assets_load)

        # Picks up asset load information from connected BCP client(s)
        self.machine.events.add_handler('assets_to_load',
                                        self._bcp_client_asset_load)

        # prevent excessive loading_assets events
        self._last_asset_event_time = None

        self._start_time = time.time()

    def get_next_id(self) -> int:
        """Return the next free id."""
        self._next_id += 1
        return self._next_id

    @property
    def loading_percent(self) -> int:
        """Return the percent of assets that are in the process of loading that have been loaded.

        This value is an integer between 0 and 100. It's reset
        when all the assets have been loaded, so it will go from 0 to 100 when
        MPF is starting up, and then go from 0 to 100 again when a mode starts,
        etc.

        Note that this percentage also includes asset loading status updates
        from a connected BCP client.
        """
        try:
            return round((self.num_assets_loaded +
                          self.num_bcp_assets_loaded) /
                         (self.num_assets_to_load +
                          self.num_bcp_assets_to_load) * 100)

        except ZeroDivisionError:   # pragma: no cover
            return 100

    # pylint: disable-msg=too-many-arguments
    def register_asset_class(self, asset_class: str, attribute: str, config_section: str,
                             disk_asset_section: str,
                             path_string: str, extensions: Iterable[str], priority: int,
                             pool_config_section: str) -> None:
        """Register a a type of assets to be controlled by the AssetManager.

        Args:
        ----
            asset_class: Reference to the class you want to register, based on
                mc.core.assets.Asset. e.g. mc.assets.images.ImageClass
            attribute: String of the name of the attribute dict that will be
                added to the main MpfMc instance. e.g. 'images' means that
                the dict of image names to image asset class instances will be
                at self.machine.images.
            config_section: String name of this assets section in the config
                files. e.g. 'images'
            disk_asset_section: String name of the section which holds settings
                for assets that are loaded from disk.
            path_string: String name of the setting from mpf-mc:paths: which
                controls the name of the folder that will hold this type of
                assets in the machine folder. e.g. 'images
            extensions: Tuple of strings, with no dots, of the types of file
                extensions that are valid for this type of asset. e.g. ('jpg',
                'gif', 'png')
            priority: Integer of the relative priority of this asset class as
                compared to other asset classes. This affects the order that
                asset objects are created and loaded (when there's a tie)
                because some asset classes depend on others to exist first.
                e.g. 'slide_shows' assets need 'images', 'videos', and 'sounds'
                to exist. Higher number is first.
            pool_config_section: String which specifies the config file
                section for associated asset groups.
        """
        if not hasattr(self.machine, attribute):
            # some assets of different classes use the same mc attribute, like
            # images and animated_images
            setattr(self.machine, attribute, dict())

        ac = AssetClass(attribute=attribute,
                        cls=asset_class,
                        path_string=path_string,
                        config_section=config_section,
                        disk_asset_section=disk_asset_section,
                        extensions=extensions,
                        priority=priority,
                        pool_config_section=pool_config_section,
                        defaults=self._get_asset_class_defaults(disk_asset_section, self.machine.config))

        self._asset_classes.append(ac)
        self._asset_classes.sort(key=lambda x: x.priority, reverse=True)

    @classmethod
    def _get_asset_class_defaults(cls, disk_asset_section: str, config) -> dict:
        # Creates the folder-based default configs for the asset class
        # starting with the default section and then created folder-specific
        # entries based on that. Just runs once on startup for each asset
        # class.
        default_config_dict = dict()

        if 'assets' in config and config['assets']:

            if (disk_asset_section in config['assets'] and
                    config['assets'][disk_asset_section]):

                this_config = config['assets'][disk_asset_section]

                # set the default
                default_config_dict['default'] = this_config['default']

                for default_section_name in this_config:
                    # first get a copy of the default for this section
                    default_config_dict[default_section_name] = (
                        copy.deepcopy(default_config_dict['default']))

                    # then merge in this section's specific settings
                    default_config_dict[default_section_name].update(
                        this_config[default_section_name])

        return default_config_dict

    def _create_assets(self, **kwargs) -> None:
        force_assets_load = kwargs.get('force_assets_load', False)

        # Called once on boot to create all the asset objects
        # Create the machine-wide assets

        self._create_assets_from_disk(config=self.machine.machine_config)
        self._create_asset_groups(config=self.machine.machine_config)

        # Create the mode assets
        for mode in self.machine.modes.values():
            self._create_assets_from_disk(config=mode.config, mode=mode)
            self._create_asset_groups(config=mode.config, mode=mode)

        # load the assets marked for preload:
        preload_assets = list()

        for ac in self._asset_classes:
            preload_assets.extend(
                [x for x in getattr(self.machine, ac.attribute).values() if
                 x.config['load'] == 'preload' or force_assets_load])

        wait_for_assets = False
        for asset in preload_assets:
            if not asset.load():
                wait_for_assets = True

        if not wait_for_assets:
            self.machine.clear_boot_hold('assets')

    def _create_assets_from_disk(self, config: dict, mode: Optional[Mode] = None) -> dict:
        """Walk a folder (and subfolders) and finds all the assets.

        Check to see if those assets have config entries in the passed config file, and
        then builds a config for each asset based on its config entry, and/or
        defaults based on the subfolder it was in or the general defaults.
        Then it creates the asset objects based on the built-up config.

        Args:
        ----
            config: A config dictionary.
            mode: Optional reference to the mode object which is used when
                  assets are being created from mode folders.

        Returns an updated config dictionary. (See the "How it works" section
        below for details.)

        Note that this method merely creates the asset object so they can be
        referenced in MPF. It does not actually load the asset files into
        memory.

        It's called on startup register machine-wide assets, and it's called
        as modes initialize (also during the startup process) to find assets
        in mode folders.

        How it works
        ============

        Every asset class that's registered with the Asset Manager has a folder
        associated it. (e.g. Images assets as associated wit the "images"
        folder.)

        This method will build a master config dict of all the assets of each
        type. It does this by walking the folder, looking for files of each
        asset type.

        When it finds a file, it checks the config to see if either (1) any
        entries exist with a name that matches the root of that file name, or
        (2) to see if there's a config for an asset with a different name
        but with a file: setting that matches this file name.

        If it finds a match, that entry's file: setting is updated with the
        default settings for assets in that folder as well as the full path to
        the file. If it does not find a match, it creates a new entry for
        that asset in the config.

        To build up the config, it will base the config on any settings
        specified in the "default" section of the "assets:" section for that
        asset class. (e.g. images will get whatever key/value pairs are in the
        assets:images:default section of the config.)

        Then it will look to see what subfolder the asset it in and if there
        are any custom default settings for that subfolder. For example, an
        image found in the /custom1 subfolder will get any settings in the
        assets:images:custom1 section of the config. These settings are merged
        into the settings from the default section.

        Finally it will merge in any settings that existed for this asset
        specifically.

        When this method is done, the config dict has been updated to include
        every asset it found in that folder and subfolders (along with its
        full path), and a config dict appropriately merged from default,
        folder-specific, and asset specific settings
        """
        if not config:      # pragma: no cover
            config = dict()

        try:
            mode_name = mode.name
            paths = mode.asset_paths
        except AttributeError:
            mode_name = None
            paths = [self.machine.machine_path]

        for ac in self._asset_classes:
            if ac.disk_asset_section not in config:
                config[ac.disk_asset_section] = dict()

            for path in paths:
                # Populate the config section for this asset class with all the
                # assets found on disk
                config[ac.disk_asset_section].update(self._create_asset_config_entries(
                    asset_class=ac,
                    config=config[ac.disk_asset_section],
                    mode_name=mode_name,
                    path=path))

            # create the actual instance of the Asset object and add it
            # to the self.machine asset attribute dict for that asset class
            for asset in config[ac.disk_asset_section]:
                if 'file' not in config[ac.disk_asset_section][asset]:      # pragma: no cover
                    msg = "The file associated with the disk-based asset '%s' declared in the " \
                          "'%s' config section could not be found" % (asset, ac.disk_asset_section)
                    self.error_log(msg)
                    raise FileNotFoundError(msg)

                getattr(self.machine, ac.attribute)[asset] = ac.cls(
                    self.machine, name=asset,
                    file=config[ac.disk_asset_section][asset]['file'],
                    config=config[ac.disk_asset_section][asset])

        return config

    # pylint: disable-msg=too-many-locals
    def _create_asset_config_entries(self, asset_class: AssetClass, config, mode_name: Optional[str] = None,
                                     path: Optional[str] = None) -> dict:
        """Scan a folder (and subfolders).

        Automatically creates or updates entries in the config dict for any asset files it finds.

        Args:
        ----
            asset_class: An asset class entry from the self._asset_classes
            dict.
            config: A dictionary which contains a list of asset names with
                settings that will be used for the specific asset. (Note this
                is not needed for all assets, as any asset file found not in
                the config dictionary will be set up with the folder it was
                found in's asset_defaults settings.)
            mode_name: Name of the mode this asset is used loaded by.
            path: A full core path to the root folder that will be searched
                for assets. This should *not* include the asset-specific path
                string. If None, machine's root folder will be searched.

        Returns: Updated config dict that has all the new assets (and their
            built-up configs) that were found on disk.

        Note that for each file found, this method will scan through the
        entire config dict to see if any entry exists for that file based on an
        entry's 'file:' setting. If it's not found, an entry is created based
        on the file name.

        Examples (based on images):

            asset_class defaults: section:
                default:
                    some_key: some_value
                foo:
                    some_key: some_value

            Based on the above asset_class defaults: section, the following
            files would get their 'defaults:' setting set to 'foo':
                /images/foo/image1.png
                /images/foo/bar/image2.png
                /images/foo/bar/bar2/image3.png

            And based on the above, the following files would get their
            'defaults:' setting set to 'default:
                /images/image4.png
                /images/foo/image5.png
                /images/other/images6.png
                /images/other/big/image7.png
        """
        if not path:
            path = self.machine.machine_path

        if not config:
            config = dict()

        root_path = os.path.join(path, asset_class.path_string)
        self.debug_log("Processing assets from base folder: %s", root_path)

        # ignore temporary files
        ignore_prefixes = (".", "~")
        # do not get fooled by windows or mac garbage
        ignore_files = ("desktop.ini", "Thumbs.db")

        # walk files in the asset root directory (include all subfolders)
        for this_path, _, files in os.walk(root_path, followlinks=True):
            # Determine the first-level sub-folder of the current path relative to the
            # asset root path.  The first level sub-folder is used to determine the
            # default asset keys based on the assets config section.
            relative_path = PurePath(this_path).relative_to(root_path)
            if relative_path.parts:
                first_level_subfolder = relative_path.parts[0]
            else:
                first_level_subfolder = None

            valid_files = [f for f in files if f.endswith(
                           asset_class.extensions) and not f.startswith(ignore_prefixes) and f != ignore_files]

            # loop over valid files in the current path
            for file_name in valid_files:
                name = os.path.splitext(file_name)[0]
                full_file_path = os.path.join(this_path, file_name)

                # determine default group based on first level sub-folder and location groups
                # configured in the assets section
                if first_level_subfolder is None or first_level_subfolder not in asset_class.defaults:
                    default_string = 'default'
                else:
                    default_string = first_level_subfolder

                # built_up_config is the built-up config dict for that will be
                # used for the entry for this asset.

                # first deepcopy the default config for this asset based
                # on its default_string (folder) since we use it as the base
                # for everything in case one of the custom folder configs
                # doesn't include all settings
                built_up_config = copy.deepcopy(
                    asset_class.defaults[default_string])

                # scan through the existing config to see if this file is used
                # as the file setting for any entry.
                found_in_config = False
                for k, v in config.items():
                    if ('file' in v and v['file'] == file_name) or name == k:
                        # if it's found, set the asset entry's name to whatever
                        # the name of this entry is
                        name = k
                        # merge in the config settings for this asset, updating
                        #  the defaults
                        built_up_config.update(config[k])
                        found_in_config = True
                        break

                # need to send the full file path to the Asset that will
                # be created so it will be able to load it later.
                built_up_config['file'] = full_file_path

                # If this asset is set to load on mode start, replace the load
                # value with one based on mode name
                if built_up_config['load'] == 'mode_start':
                    built_up_config['load'] = '{}_start'.format(mode_name)

                # Update the config for that asset

                if name in config and not found_in_config:      # pragma: no cover
                    raise RuntimeError(
                        "Duplicate Asset name found: {}".format(name))

                config[name] = built_up_config

                self.debug_log("Registering Asset: %s, File: %s, Default "
                               "Group: %s, Final Config: %s", name, file_name,
                               default_string, built_up_config)
        return config

    def _create_asset_groups(self, config, mode=None) -> None:
        # creates named groups of assets and adds them to to the mc's asset
        # dicts
        del mode
        for ac in [x for x in self._asset_classes
                   if x.pool_config_section]:

            if ac.pool_config_section in config:
                for name, settings in config[ac.pool_config_section].items():
                    getattr(self.machine, ac.attribute)[name] = (
                        ac.cls.asset_group_class(self.machine, name, settings,
                                                 ac.cls))

    def _load_mode_assets(self, config, priority: int, mode: Mode) -> \
            Tuple[Callable[[Iterable["Asset"]], None], Set["Asset"]]:
        # Called on mode start to load the assets that are set to automatically
        # load based on that mode starting
        del config
        return (self.unload_assets,
                self.load_assets_by_load_key(
                    key_name='{}_start'.format(mode.name),
                    priority=priority))

    def load_assets_by_load_key(self, key_name: str, priority: int = 0) -> Set["Asset"]:
        """Load all the assets with a given load key.

        Args:
        ----
            key_name: String of the load: key name.
            priority: Priority of this asset.
        """
        del priority
        assets = set()
        # loop through all the registered assets of each class and look for
        # this key name
        for ac in self._asset_classes:
            asset_objects = getattr(self.machine, ac.attribute).values()
            for asset in [x for x in asset_objects if
                          x.config['load'] == key_name]:
                asset.load()
                assets.add(asset)

        return assets

    @classmethod
    def unload_assets(cls, assets: Iterable["Asset"]) -> None:
        """Unload multiple assets.

        Args:
        ----
            assets: An iterable of asset objects. You can safely mix
                    different classes of assets.
        """
        for asset in assets:
            asset.unload()

    def load_asset(self, asset: "Asset") -> None:
        """Load an asset."""
        raise NotImplementedError("implement")

    def _bcp_client_asset_load(self, total, remaining, **kwargs):
        # Callback for the BCP assets_to_load command which tracks asset
        # loading from a connected BCP client.
        del kwargs
        self.num_bcp_assets_loaded = int(total) - int(remaining)
        self.num_bcp_assets_to_load = int(total)
        self._post_loading_event()

    def _post_loading_event(self):
        # Called each time an asset is loaded.
        total = self.num_assets_to_load + self.num_bcp_assets_to_load
        remaining = total - self.num_assets_loaded - self.num_bcp_assets_loaded

        # limit loading_assets events to max 5 per second
        if remaining and self._last_asset_event_time and \
                self._last_asset_event_time > self.machine.clock.get_time() - 0.2:
            return

        self._last_asset_event_time = self.machine.clock.get_time()

        self.machine.events.post(
            'loading_assets', total=total,
            loaded=self.num_assets_loaded + self.num_bcp_assets_loaded,
            remaining=remaining,
            percent=self.loading_percent)
        '''event: loading_assets

        desc: Posted when the number of assets waiting to be loaded changes.

        Note that once all the assets are loaded, all the values below are
        reset to zero.

        args:

        total: The total number of assets that need to be loaded. This is
        equal to the sum of the *loaded* and *remaining* values below. It
        also includes assets that MPF is loading itself as well as any
        assets that have been reported from remotely connected BCP hosts
        (e.g. the media controller).

        loaded: The number of assets that have been loaded so far.

        remaining: The number of assets that are remaining to be loaded.

        percent: The numerical percent completion of the assets loaded, express
            in the range of 0 to 100.

        '''

        if not remaining:
            if self._start_time:
                self.log.info("Asset loading took: %s", time.time() - self._start_time)
            self._last_asset_event_time = None
            self.machine.events.post('asset_loading_complete')
            '''event: asset_loading_complete
            desc: Posted when the asset manager has loaded all the assets in
            its queue.

            Note that this event does *NOT* necessarily mean that all asset
            loading is complete. Rather is just means that the asset manager
            has loaded everything in its queue.

            For example, when the MPF-MC boots, it will load the assets it is
            configured to load on start. However, if the MPF MC is started but
            MPF is not, then the MPF MC will load its assets and then post this
            *asset_loading_complete* event when it's done. Then when MPF is
            started and connects, MPF will need to load its own assets, which
            means the MPF MC will post more *loading_assets* and then
            a final *asset_loading_complete* event a second time for the
            MPF-based assets.
            '''

        self.info_log('Loading assets: %s/%s (%s%%)',
                      self.num_assets_loaded + self.num_bcp_assets_loaded,
                      total, self.loading_percent)

        if not remaining and not self.machine.is_init_done.is_set():
            self.machine.clear_boot_hold('assets')
            self.initial_assets_loaded = True


class AsyncioSyncAssetManager(BaseAssetManager):

    """AssetManager which uses asyncio to load assets."""

    __slots__ = []

    @staticmethod
    def _load_sync(asset):
        if not asset.loaded:
            asset.do_load()
            return True

        return False

    async def wait_for_asset_load(self, asset):
        """Wait for an asset to load."""
        result = self._load_sync(asset)
        if result:
            asset.is_loaded()
        self.num_assets_loaded += 1
        self._post_loading_event()

    def load_asset(self, asset):
        """Load an asset."""
        self.num_assets_to_load += 1
        task = self.machine.clock.loop.create_task(self.wait_for_asset_load(asset))
        task.add_done_callback(Util.raise_exceptions)


TUnconditionalAssetEntry = Tuple[AssetClass, int]
TConditionalAssetEntry = Tuple[AssetClass, int, BoolTemplate]
TMaybeConditionalAssetEntry = Union[TUnconditionalAssetEntry, TConditionalAssetEntry]
TNullableAssetEntry = Union[TMaybeConditionalAssetEntry, Tuple[None, int]]


# pylint: disable=too-many-instance-attributes
class AssetPool:

    """Pool of assets."""

    __slots__ = ["machine", "priority", "name", "config", "member_cls", "loading_members", "_callbacks", "assets",
                 "_last_asset", "_asset_sequence", "_assets_sent", "_total_weights", "_has_conditions"]

    # Could possibly combine some or make @properties?
    def __init__(self, mc, name, config, member_cls):
        """Initialise asset pool."""
        self.machine = mc
        self.priority = None
        self.name = name
        self.config = config
        self.member_cls = member_cls
        self.loading_members = set()
        self._callbacks = set()
        self.assets = list()         # type: List[TMaybeConditionalAssetEntry]
        self._last_asset = None
        self._asset_sequence = deque()
        self._assets_sent = set()
        self._total_weights = 0
        self._has_conditions = False

        if 'load' not in config:
            config['load'] = 'on_demand'

        if 'type' not in config:
            config['type'] = 'sequence'

        for asset in Util.string_to_event_list(self.config[self.member_cls.config_section]):
            asset_condition = self.machine.placeholder_manager.parse_conditional_template(asset, default_number=1)

            # For efficiency, track whether any assets have conditions
            if asset_condition.condition:
                self._has_conditions = True

            try:
                self.assets.append((
                    getattr(self.machine, self.member_cls.attribute)[asset_condition.name],
                    asset_condition.number, asset_condition.condition))
            except KeyError:    # pragma: no cover
                raise ValueError("No asset named {}".format(asset_condition.name))

        self._configure_return_asset()

    def __repr__(self):
        """Return string representation."""
        return '<AssetPool: {}>'.format(self.name)

    @property
    def asset(self):
        """Pop one asset from the pool."""
        return self.get_next()

    def get_next(self):
        """Pop one asset from the pool."""
        if self.config['type'] == 'random':
            return self._get_random_asset()
        if self.config['type'] == 'sequence':
            return self._get_sequence_asset()
        if self.config['type'] == 'random_force_next':
            return self._get_random_force_next_asset()
        if self.config['type'] == 'random_force_all':
            return self._get_random_force_all_asset()

        raise AssertionError("Invalid type {}".format(self.config['type']))

    @property
    def loaded(self):
        """Return if loaded."""
        for asset in self.assets:
            if not asset[0].loaded:
                return False

        return True

    @property
    def loading(self):
        """Return if loading."""
        for asset in self.assets:
            if asset[0].loading:
                return True

        return False

    @property
    def unloading(self):
        """Return if unloading."""
        for asset in self.assets:
            if asset[0].unloading:
                return True

        return False

    def _configure_return_asset(self):
        self._total_weights = sum([x[1] for x in self.assets])

        if self.config['type'] == 'sequence':
            for index in range(len(self.assets)):
                self._asset_sequence.extend([self.assets[index][0]] *
                                            self.assets[index][1])
            self._asset_sequence.rotate(1)

    def load(self, callback=None, priority=None) -> bool:
        """Load all assets in the pool."""
        # No need to attempt to load the asset if it is already loading
        if priority is not None:
            self.priority = priority

        if callback:
            self._callbacks.add(callback)

        for asset in self.assets:
            if not asset[0].loaded and not asset[0].loading:
                self.loading_members.add(asset)
                asset[0].load(callback=self._group_member_loaded)

        if not self.loading_members:
            self._call_callbacks()
            return True

        return False

    def unload(self):
        """Unload all assets in the pool."""
        for asset in self.assets:
            if asset[0].loaded:
                asset[0].unload()

    def _group_member_loaded(self, asset):
        self.loading_members.discard(asset)
        if not self.loading_members:
            self._call_callbacks()

    def _call_callbacks(self):
        for callback in self._callbacks:

            if callable(callback):
                callback(self)

        self._callbacks = set()

    def _get_conditional_assets(self) -> List[TNullableAssetEntry]:
        if not self._has_conditions:
            return self.assets

        result = [asset for asset in self.assets
                  if not asset[2] or asset[2].evaluate([])]  # type: List[TNullableAssetEntry]
        # Avoid crashes, return None as the asset if no conditions evaluate true
        if not result:
            self.machine.log.warning("AssetPool {}: {}".format(
                self.name, "All conditional assets evaluated False and no other assets defined."))
            result.append((None, 0))
        return result

    def _get_random_asset(self) -> AssetClass:
        conditional_assets = self._get_conditional_assets()  # Store to variable to avoid calling twice
        # If any asset in the group has a condition, recalculate the total weight based on what's valid
        if self._has_conditions:
            self._total_weights = sum([x[1] for x in conditional_assets])
        return self._pick_weighed_random(conditional_assets)[0]

    def _get_sequence_asset(self) -> Optional[AssetClass]:
        self._asset_sequence.rotate(-1)
        if self._has_conditions:
            # Get the names of all assets that evaluate true
            truthy_asset_names = [asset[0].name for asset in self._get_conditional_assets()]
            # Skip any assets that have falsey conditions
            for x in range(len(self._asset_sequence)):
                if self._asset_sequence[0].name in truthy_asset_names:
                    break
                if x == len(self._asset_sequence) - 1:
                    self.machine.log.warning("AssetPool {}: All assets in sequence evaluated False.".format(self.name))
                    return None

                self._asset_sequence.rotate(-1)
        return self._asset_sequence[0]

    def _get_random_force_next_asset(self) -> Optional[AssetClass]:
        conditional_assets = self._get_conditional_assets()  # Store to variable to avoid calling twice
        self._total_weights = sum([x[1] for x in conditional_assets
                                   if x is not self._last_asset])

        self._last_asset = self._pick_weighed_random([x for x in conditional_assets
                                                      if
                                                      x is not
                                                      self._last_asset])
        return self._last_asset[0]

    def _get_random_force_all_asset(self) -> Optional[AssetClass]:
        conditional_assets = self._get_conditional_assets()  # Store to variable to avoid calling twice
        if len(self._assets_sent) == len(conditional_assets):
            self._assets_sent = set()

        asset = self._pick_weighed_random([x for x in conditional_assets
                                           if x not in self._assets_sent])
        if asset[0] is not None:
            self._assets_sent.add(asset)
        return asset[0]

    def _pick_weighed_random(self, assets: List[TConditionalAssetEntry]) -> TNullableAssetEntry:
        if not assets or assets[0][0] is None:
            return (None, 0)

        value = random.randint(1, self._total_weights)  # nosec
        index_value = assets[0][1]

        for asset in assets:
            if index_value >= value:
                return asset

            index_value += asset[1]

        return assets[-1]

    @property
    def is_pool(self):
        """Return that this is a pool."""
        return True


class Asset:

    """Baseclass for all assets."""

    attribute = ''  # attribute in MC, e.g. self.machine.images
    path_string = ''  # entry from mpf-mc:paths: for asset folder name
    config_section = ''  # section in the config files for this asset
    disk_asset_section = ''  # option is assets: config name is different

    # tuple of strings, no dots
    extensions = tuple()  # type: Iterable[str]
    class_priority = 0  # Order asset classes will be loaded. Higher is first.

    # Create an associated AssetPool instance
    pool_config_section = None  # type: str

    asset_group_class = AssetPool  # replace with your own asset group class

    __slots__ = ["machine", "name", "file", "config", "priority", "_callbacks", "_id", "lock", "loading", "loaded",
                 "unloading"]

    @classmethod
    def initialize(cls, machine):
        """Initialise asset class."""
        if not cls.disk_asset_section:
            cls.disk_asset_section = cls.config_section

        machine.asset_manager.register_asset_class(
            asset_class=cls,
            attribute=cls.attribute,
            path_string=cls.path_string,
            config_section=cls.config_section,
            disk_asset_section=cls.disk_asset_section,
            extensions=cls.extensions,
            priority=cls.class_priority,
            pool_config_section=cls.pool_config_section)

    def __init__(self, machine, name, file, config):
        """Initialise asset."""
        self.machine = machine      # type: MachineController
        self.name = name
        self.file = file

        self.config = self.machine.config_validator.validate_config(
            'assets:{}'.format(self.config_section), config,
            base_spec='assets:common')

        self.priority = self.config.get('priority', 0)
        self._callbacks = set()
        self._id = machine.asset_manager.get_next_id()
        self.lock = threading.Lock()

        self.loading = False  # Is this asset in the process of loading?
        self.loaded = False  # Is this asset loaded and ready to use?
        self.unloading = False  # Is this asset in the process of unloading?

    def __repr__(self):
        """Return string representation."""
        return '<{} Asset: {}, loaded={}>'.format(self.attribute.capitalize(),
                                                  self.name, self.loaded)

    def __lt__(self, other):
        """Compare assets."""
        # Note this is "backwards" (It's the __lt__ method but the formula uses
        # greater than because the PriorityQueue puts lowest first.)
        return ("%s, %s" % (self.priority, self._id) >
                "%s, %s" % (other.priority, other.get_id()))

    def get_id(self):
        """Return id."""
        return self._id

    def load(self, callback=None, priority=None) -> bool:
        """Start loading the asset.

        Returns True if the asset is already loaded.
        """
        # No need to attempt to load the asset if it is already loading
        if self.loading:
            # Add the supplied callback before returning
            if callback:
                self._callbacks.add(callback)
            return False

        if priority is not None:
            self.priority = priority

        if callback:
            self._callbacks.add(callback)

        if self.loaded:
            self._call_callbacks()
            return True

        if self.unloading:
            pass
            # do something fancy here. Maybe just skip it and come back?

        self.loading = True
        self.machine.asset_manager.load_asset(self)
        return False

    def _call_callbacks(self):
        for callback in self._callbacks:
            if callable(callback):
                callback(self)

        self._callbacks = set()

    def do_load(self):
        """Load the asset blocking."""
        # This is the actual method that loads the asset. It's called by a
        # different thread so it's ok to block. Make sure you don't set any
        # attributes here or you don't need any since it's a separate thread.
        raise NotImplementedError

    def is_loaded(self):
        """Handle that asset has been loaded."""
        self._call_callbacks()
        self.loading = False
        self.loaded = True
        self.unloading = False

    def unload(self):
        """Handle that asset has been unloaded."""
        self.unloading = True
        self.loaded = False
        self.loading = False
        self._do_unload()
        self.unloading = False

    def _do_unload(self):
        # This is the actual method that unloads the asset
        raise NotImplementedError

    @property
    def is_pool(self):
        """Return that this is not a pool."""
        return False



"""Test multiball_locks."""
from mpf.tests.MpfGameTestCase import MpfGameTestCase
from mpf.tests.MpfTestCase import test_config


class TestMultiballLock(MpfGameTestCase):
    def get_config_file(self):
        return 'testDefault.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/multiball_locks/'

    def get_platform(self):
        return 'smart_virtual'

    @test_config("testSourceDevices.yaml")
    def test_source_devices(self):
        self.fill_troughs()
        self.start_game()
        self.post_event("start_source_devices")
        self.assertModeRunning("source_devices")

        lock_device = self.machine.ball_devices["bd_lock_triple"]
        lock_device2 = self.machine.ball_devices["bd_lock"]
        self.machine.default_platform.add_ball_to_device(lock_device)
        self.advance_time_and_run(10)
        self.assertEqual(1, lock_device.balls)
        self.assertEqual(0, lock_device2.balls)
        self.assertBallsInPlay(1)
        self.assertBallsOnPlayfield(1)

        self.machine.default_platform.add_ball_to_device(lock_device2)
        self.advance_time_and_run(10)
        self.assertEqual(0, lock_device.balls)
        self.assertEqual(1, lock_device2.balls)
        self.assertBallsInPlay(1)
        self.assertBallsOnPlayfield(1)

        self.machine.default_platform.add_ball_to_device(lock_device2)
        self.advance_time_and_run(10)
        self.assertEqual(0, lock_device.balls)
        self.assertEqual(2, lock_device2.balls)
        self.assertBallsInPlay(1)
        self.assertBallsOnPlayfield(1)

    def test_filling_two(self):
        self.fill_troughs()
        self.start_game()
        self.mock_event("multiball_lock_lock_default_locked_ball")
        self.mock_event("multiball_lock_lock_default_full")
        self.post_event("start_default")

        lock_device = self.machine.ball_devices["bd_lock"]
        mb_lock = self.machine.multiball_locks["lock_default"]

        # Add one ball, should lock but not be full
        self.machine.default_platform.add_ball_to_device(lock_device)
        self.advance_time_and_run(10)
        self.assertEqual(1, lock_device.balls)
        self.assertEqual(1, mb_lock.locked_balls)
        self.assertEventCalled("multiball_lock_lock_default_locked_ball")
        self.assertEqual({'total_balls_locked': 1},
                         self._last_event_kwargs["multiball_lock_lock_default_locked_ball"])
        self.assertEventNotCalled("multiball_lock_lock_default_full")

        # Add a second ball, should lock and be full
        self.machine.default_platform.add_ball_to_device(lock_device)
        self.advance_time_and_run(10)
        self.assertEqual(2, lock_device.balls)
        self.assertEqual(2, mb_lock.locked_balls)
        self.assertEventCalled("multiball_lock_lock_default_locked_ball")
        self.assertEqual({'total_balls_locked': 2},
                         self._last_event_kwargs["multiball_lock_lock_default_locked_ball"])
        self.assertEventCalledWith("multiball_lock_lock_default_full", balls=2)

    def test_filling_three(self):
        self.fill_troughs()
        self.start_game()
        self.mock_event("multiball_lock_lock_triple_locked_ball")
        self.mock_event("multiball_lock_lock_triple_full")
        self.post_event("start_default")

        lock_device_3 = self.machine.ball_devices["bd_lock_triple"]
        mb_lock_3 = self.machine.multiball_locks["lock_triple"]

        # Add one ball, should lock but not be full
        self.machine.default_platform.add_ball_to_device(lock_device_3)
        self.advance_time_and_run(10)
        self.assertEqual(1, lock_device_3.balls)
        self.assertEqual(1, mb_lock_3.locked_balls)
        self.assertEventCalled("multiball_lock_lock_triple_locked_ball")
        self.assertEqual({'total_balls_locked': 1},
                         self._last_event_kwargs["multiball_lock_lock_triple_locked_ball"])
        self.assertEventNotCalled("multiball_lock_lock_triple_full")

        # Add a second ball, should lock but not be full
        self.machine.default_platform.add_ball_to_device(lock_device_3)
        self.advance_time_and_run(10)
        self.assertEqual(2, lock_device_3.balls)
        self.assertEqual(2, mb_lock_3.locked_balls)
        self.assertEventCalled("multiball_lock_lock_triple_locked_ball")
        self.assertEqual({'total_balls_locked': 2},
                         self._last_event_kwargs["multiball_lock_lock_triple_locked_ball"])
        self.assertEventNotCalled("multiball_lock_lock_triple_full")

        # Add a third ball, should lock but not be full
        self.machine.default_platform.add_ball_to_device(lock_device_3)
        self.advance_time_and_run(10)
        self.assertEqual(3, lock_device_3.balls)
        self.assertEqual(3, mb_lock_3.locked_balls)
        self.assertEventCalled("multiball_lock_lock_triple_locked_ball")
        self.assertEqual({'total_balls_locked': 3},
                         self._last_event_kwargs["multiball_lock_lock_triple_locked_ball"])
        self.assertEventCalledWith("multiball_lock_lock_triple_full", balls=3)

    def test_placeholder_events(self):
        self.fill_troughs()
        self.start_game()

        self.post_event("start_default")
        self.mock_event("should_post_when_enabled")
        self.mock_event("should_post_when_disabled")
        self.mock_event("should_not_post_when_enabled")
        self.mock_event("should_not_post_when_disabled")
        lock = self.machine.multiball_locks["lock_default"]

        self.assertTrue(lock.enabled)
        self.assertPlaceholderEvaluates(True, "device.multiball_locks.lock_default.enabled")
        self.post_event("test_event_when_enabled")
        self.assertEventCalled("should_post_when_enabled")
        self.assertEventNotCalled("should_not_post_when_enabled")

        lock.disable()
        self.assertFalse(lock.enabled)
        self.assertPlaceholderEvaluates(False, "device.multiball_locks.lock_default.enabled")
        self.post_event("test_event_when_disabled")
        self.assertEventCalled("should_post_when_disabled")
        self.assertEventNotCalled("should_not_post_when_disabled")


class TestMultiballLockCountingStrategies(MpfGameTestCase):

    def get_config_file(self):
        return self._testMethodName + '.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/multiball_locks/'

    def get_platform(self):
        return 'smart_virtual'

    def testNoVirtual(self):
        # prepare game
        self.fill_troughs()

        # start game
        self.start_two_player_game()

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        self.advance_time_and_run(4)
        self.assertEqual(1, self.machine.playfield.available_balls)

        lock = self.machine.multiball_locks["lock_no_virtual"]

        # start mode
        self.post_event("start_no_virtual")

        self.advance_time_and_run(4)
        self.assertEqual(0, lock.locked_balls)

        # lock one ball and another one should go to pf
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_no_virtual")

        self.assertPlayerNumber(2)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(0, lock.locked_balls)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)
        self.assertEqual(0, lock.locked_balls)

        # lock ejects one ball
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_no_virtual")
        self.assertPlayerNumber(1)
        self.assertEqual(0, lock.locked_balls)

        # lock one ball
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)

        # it should not keep the ball
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_no_virtual")
        self.assertPlayerNumber(2)
        self.assertEqual(0, lock.locked_balls)

        # lock one ball
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)

        # it should not keep the ball
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        # lock a second ball
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)

        # it should keep two balls
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)

        # game ends
        self.drain_one_ball()
        self.advance_time_and_run(10)
        self.assertGameIsNotRunning()

        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(2, self.machine.playfield.balls)

        # game should not start yet
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsNotRunning()

        self.drain_one_ball()
        self.drain_one_ball()
        self.advance_time_and_run()

        # game should start again
        self.start_game()

    def testVirtualOnly(self):
        # prepare game
        self.fill_troughs()

        # start game
        self.start_two_player_game()

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        self.advance_time_and_run(4)
        self.assertEqual(1, self.machine.playfield.available_balls)

        # start mode
        self.post_event("start_virtual_only")

        lock = self.machine.multiball_locks["lock_virtual_only"]

        self.advance_time_and_run(4)
        self.assertEqual(0, lock.locked_balls)

        # lock one ball and another one should go to pf
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_virtual_only")

        self.assertPlayerNumber(2)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(0, lock.locked_balls)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # lock ejects one ball
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_virtual_only")
        self.assertPlayerNumber(1)
        self.assertEqual(1, lock.locked_balls)

        # lock one ball
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)

        # it should keep the ball
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_virtual_only")
        self.assertPlayerNumber(2)

        # lock stays full
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)

        # game ends
        self.drain_one_ball()
        self.advance_time_and_run(10)
        self.assertGameIsNotRunning()

        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(2, self.machine.playfield.balls)

        # game should not start yet
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsNotRunning()

        self.drain_one_ball()
        self.drain_one_ball()
        self.advance_time_and_run()

        # game should start again
        self.start_game()

    def testPhysicalOnly(self):
        self.mock_event("multiball_lock_lock_physical_only_locked_ball")
        self.mock_event("multiball_lock_lock_physical_only_full")
        # prepare game
        self.fill_troughs()

        # start game
        self.start_two_player_game()

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        self.advance_time_and_run(4)
        self.assertEqual(1, self.machine.playfield.available_balls)

        # start mode
        self.post_event("start_physical_only")

        self.mock_event("multiball_lock_lock_physical_only_smaller_than_device_locked_ball")
        self.mock_event("multiball_lock_lock_physical_only_smaller_than_device_full")
        lock2 = self.machine.multiball_locks["lock_physical_only_smaller_than_device"]
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock_triple"])
        self.advance_time_and_run(10)

        self.assertEqual(1, self.machine.ball_devices["bd_lock_triple"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock2.locked_balls)
        self.assertEventNotCalled("multiball_lock_lock_physical_only_smaller_than_device_full")
        self.assertEventCalled("multiball_lock_lock_physical_only_smaller_than_device_locked_ball", times=1)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock_triple"])
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_devices["bd_lock_triple"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock2.locked_balls)
        self.assertEventCalled("multiball_lock_lock_physical_only_smaller_than_device_full")
        self.assertEventCalled("multiball_lock_lock_physical_only_smaller_than_device_locked_ball", times=2)

        self.mock_event("multiball_lock_lock_physical_only_smaller_than_device_full")
        self.mock_event("multiball_lock_lock_physical_only_smaller_than_device_locked_ball")
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock_triple"])
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_devices["bd_lock_triple"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock2.locked_balls)
        self.assertEventNotCalled("multiball_lock_lock_physical_only_smaller_than_device_full")
        self.assertEventNotCalled("multiball_lock_lock_physical_only_smaller_than_device_locked_ball")

        lock = self.machine.multiball_locks["lock_physical_only"]

        self.advance_time_and_run(4)
        self.assertEqual(0, lock.locked_balls)

        # lock one ball and another one should go to pf
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)
        self.assertEventCalledWith("multiball_lock_lock_physical_only_locked_ball", total_balls_locked=1)
        self.assertEventNotCalled("multiball_lock_lock_physical_only_full")

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.mock_event("multiball_lock_lock_physical_only_locked_ball")
        self.post_event("start_physical_only")

        self.assertPlayerNumber(2)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)
        self.assertEventCalledWith("multiball_lock_lock_physical_only_locked_ball", total_balls_locked=2)
        self.assertEventCalledWith("multiball_lock_lock_physical_only_full", balls=2)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_physical_only")
        self.assertPlayerNumber(1)

        # look is still full
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)
        # eject a ball
        self.machine.ball_devices["bd_lock"].eject()
        self.advance_time_and_run(10)

        # ball is on playfield now
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_physical_only")
        self.assertPlayerNumber(2)

        # only one ball locked. other player stole it
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        # game ends
        self.drain_one_ball()
        self.advance_time_and_run(10)
        self.assertGameIsNotRunning()

        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(4, self.machine.playfield.balls)

        # game should not start yet
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsNotRunning()

        self.drain_one_ball()
        self.drain_one_ball()
        self.drain_one_ball()
        self.drain_one_ball()
        self.advance_time_and_run()

        # game should start again
        self.start_game()

    def testMinVirtualPhysical(self):
        # prepare game
        self.fill_troughs()

        # start game
        self.start_two_player_game()

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        self.advance_time_and_run(4)
        self.assertEqual(1, self.machine.playfield.available_balls)

        # start mode
        self.post_event("start_min_virtual_physical")

        lock = self.machine.multiball_locks["lock_min_virtual_physical"]

        self.advance_time_and_run(4)
        self.assertEqual(0, lock.locked_balls)

        # lock one ball and another one should go to pf
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # lock ejects one ball
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_min_virtual_physical")

        self.assertPlayerNumber(2)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(0, lock.locked_balls)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)

        # steal one ball
        self.machine.ball_devices["bd_lock"].eject()
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEqual(1, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_min_virtual_physical")
        self.assertPlayerNumber(1)
        # the other player stole a ball
        self.assertEqual(1, lock.locked_balls)

        # lock one ball
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(10)

        # it should keep the ball
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)

        # player change
        self.drain_one_ball()
        self.advance_time_and_run(10)

        # start mode
        self.post_event("start_min_virtual_physical")
        self.assertPlayerNumber(2)

        # it should keep the ball
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEqual(2, lock.locked_balls)
        # game ends
        self.drain_one_ball()
        self.advance_time_and_run(10)
        self.assertGameIsNotRunning()

        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(3, self.machine.playfield.balls)

        # game should not start yet
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsNotRunning()

        self.drain_one_ball()
        self.drain_one_ball()
        self.drain_one_ball()
        self.advance_time_and_run()

        # game should start again
        self.start_game()



from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallDeviceRouting(MpfTestCase):
    def __init__(self, test_map):
        super().__init__(test_map)
        self._captured = 0
        self._missing = 0

    def get_config_file(self):
        return 'test_ball_device_routing.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def _missing_ball(self, **kwargs):
        del kwargs
        self._missing += 1

    def _captured_from_pf(self, balls, **kwargs):
        del kwargs
        self._captured += balls

    def test_routing_to_pf_on_capture(self):
        c_launcher = self.machine.coils['c_launcher']
        c_launcher.pulse = MagicMock()
        trough1 = self.machine.ball_devices['test_trough1']
        trough2 = self.machine.ball_devices['test_trough2']
        launcher = self.machine.ball_devices['test_launcher']
        target1 = self.machine.ball_devices['test_target1']

        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)

        self._missing = 0
        self._captured = 0

        self.machine.switch_controller.process_switch("s_launcher", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)
        self.assertEqual(1, self._captured)
        self._captured = 0

        self.assertTrue(c_launcher.pulse.called)
        self.assertIsNone(trough1.outgoing_balls_handler._current_target)
        self.assertIsNone(target1.outgoing_balls_handler._current_target)
        self.assertEqual(trough2, launcher.outgoing_balls_handler._current_target)
        self.assertEqual("ejecting", launcher._state)

        self.machine.switch_controller.process_switch("s_launcher", 0)
        self.advance_time_and_run(1)

        self.assertEqual(0, self._missing)

    def test_drain_routing(self):
        trough2 = self.machine.ball_devices['test_trough2']
        drain1 = self.machine.ball_devices['test_drain']
        c_drain1 = self.machine.coils['c_drain1']
        c_drain1.pulse = MagicMock()

        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)

        self._missing = 0
        self._captured = 0

        self.machine.switch_controller.process_switch("s_drain1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, drain1.balls)
        self.assertEqual(1, self._captured)
        self._captured = 0
        self.assertTrue(c_drain1.pulse.called)

        self.assertIsNone(trough2.outgoing_balls_handler._current_target)
        self.assertEqual("ejecting", drain1._state)
        self.assertEqual(trough2, drain1.outgoing_balls_handler._current_target)

        self.machine.switch_controller.process_switch("s_drain1", 0)
        self.advance_time_and_run(1)

        self.assertEqual(0, self._missing)



"""Segment displays on light drivers."""
import logging
from mpf.core.segment_mappings import SEVEN_SEGMENTS, BCD_SEGMENTS, FOURTEEN_SEGMENTS, SIXTEEN_SEGMENTS
from mpf.platforms.interfaces.segment_display_platform_interface import SegmentDisplaySoftwareFlashPlatformInterface
from mpf.core.platform import SegmentDisplaySoftwareFlashPlatform


class LightSegmentDisplay(SegmentDisplaySoftwareFlashPlatformInterface):

    """Segment display which drives lights."""

    __slots__ = ["_lights", "_key", "_segment_map"]

    def __init__(self, number, lights, segment_type):
        """Initialise segment display."""
        super().__init__(number)
        self._lights = lights
        if segment_type == "7segment":
            self._segment_map = SEVEN_SEGMENTS
        elif segment_type == "bcd":
            self._segment_map = BCD_SEGMENTS
        elif segment_type == "14segment":
            self._segment_map = FOURTEEN_SEGMENTS
        elif segment_type == "16segment":
            self._segment_map = SIXTEEN_SEGMENTS
        else:
            raise AssertionError("Invalid segment type {}".format(segment_type))

        self._key = "segment_display_{}".format(number)

    def _set_text(self, text: str) -> None:
        """Set text to lights."""
        # get the last chars for the number of chars we have
        text = text[-len(self._lights):]
        text = text.zfill(len(self._lights))
        # iterate lights and chars
        for char, lights_for_char in zip(text, self._lights):
            try:
                char_map = self._segment_map[ord(char)]
            except KeyError:
                # if there is no
                char_map = self._segment_map[None]
            for name, light in lights_for_char.items():
                if getattr(char_map, name):
                    light.on(key=self._key)
                else:
                    light.remove_from_stack_by_key(key=self._key)


class LightSegmentDisplaysPlatform(SegmentDisplaySoftwareFlashPlatform):

    """Platform which drives segment displays on lights of another platform."""

    __slots__ = ["log", "config"]

    def __init__(self, machine):
        """Initialise platform."""
        super().__init__(machine)
        self.log = logging.getLogger('Light Segment Displays')
        self.log.debug("Configuring Light Segment Displays")
        self.config = self.machine.config_validator.validate_config("light_segment_displays",
                                                                    self.machine.config.get("light_segment_displays"))

    async def configure_segment_display(self, number: str, platform_settings) -> LightSegmentDisplay:
        """Configure light segment display."""
        settings = self.machine.config_validator.validate_config("light_segment_displays_device", platform_settings)
        display = LightSegmentDisplay(number, lights=settings['lights'], segment_type=settings['type'])
        self._handle_software_flash(display)
        return display



"""Spike defines."""
from typing import List


class SpikeNodebus:

    """Spike message defines."""

    __slots__ = []  # type: List[str]

    Poll = 0x00
    GetBridgeVersion = 0x03
    GetBridgeStatus = 0x05
    SetResponseTime = 0x06
    SetPower = 0x07             # Spike 2
    SetAmpPower = 0x08          # Spike 2
    SetBackboxLight = 0x09      # Spike 2
    GetBridgeState = 0x0a
    SetBackboxPWMFreqDiv = 0x0b     # Spike 2
    GetInputState = 0x11
    GetInputMask = 0x12         # unused
    GetLEDMask = 0x13           # Spike 2
    SetNumLEDsInputs = 0x14     # Spike 2
    SSPWrite = 0x20             # unused
    SSPRead = 0x21              # unused
    StepperStatus = 0x30        # unused
    StepperSet = 0x31           # unused
    StepperConfig = 0x32        # unused
    StepperHome = 0x34          # unused
    StepperInfo = 0x38          # unused
    CoilFireRelease = 0x40
    CoilSetReflex = 0x41
    CoilGetPWMFreq = 0x42       # unused; len: 1?
    CoilSetPriority = 0x43      # used in newer firmwares
    CoilSetOCIgnoreMap = 0x44   # Spike 1
    CoilSetOCBehavior = 0x44    # Spike 2; Same as CoilSetOCIgnoreMap
    CoilShiftData = 0x45        # Spike 2; Uses SendData (F4)
    CoilSetMask = 0x46          # used during oc detection. two bytes of mask
    CoilSetOCTime = 0x48        # used during oc detection. two bytes of time
    OCGetFault = 0x4a           # Spike 2
    SetAC = 0x50                # unused; len: 4
    MotorConfigure = 0x51       # unused; len: 12; 10 bytes of motor config
    MotorStatus = 0x52          # unused
    MotorHome = 0x53            # unused; len: 6
    MotorAway = 0x54            # unused; len: 6
    MotorGo = 0x55              # unused; len: 8
    MotorStop = 0x56            # Spike 2
    MotorConfigure2 = 0x57      # Spike 2; NODEBUS_DoMotorConfigure uses this when param4 is set
    ConfigQuadrature = 0x60     # unused
    ConfigInput = 0x70          # unused
    SetLEDMask = 0x72           # Spike 2
    SetSyncTime = 0x73          # Spike 2
    SetLed = 0x80               # can set one or multiple LEDs
    SetRGBMulti = 0xC0          # Spike 2; name unknown; part of NODEBUS_SetLEDMultiple when LED > 0x40
    SetRGB = 0xC1               # Spike 2
    RGBSuspend = 0xC3           # Spike 2
    RGBResume = 0xC4            # Spike 2
    SetRGBPalette = 0xCF        # Spike 2
    SetRGBSweep = 0xC2          # Spike 2
    SetTraffic = 0xF0
    Reset = 0xF1
    LCDSet = 0xF2               # unused
    SendKey = 0xF3              # does something with CRC32
    SendData = 0xF4             # unused; not sure
    GetChecksum = 0xF5          # verify firmware somehow
    GetDebug = 0xF6             # unused; maybe also EraseFlash
    WriteFlash = 0xF7           # unused; maybe also wrong
    GetBootStatus = 0xF8        # unused
    GetFullBoardId = 0xF9       # unused; len: 2, reads 0x10 bytes back
    GetCoilCurrent = 0xFA
    DisableOC = 0xFB            # unused
    GetBoardId = 0xFD           # unused; len: 1, probably reads 0x0C bytes back
    GetVersion = 0xFE
    GetStatus = 0xFF



"""A generic state machine."""
from mpf.core.device_monitor import DeviceMonitor

from mpf.core.mode import Mode
from mpf.core.player import Player
from mpf.core.mode_device import ModeDevice
from mpf.core.system_wide_device import SystemWideDevice


@DeviceMonitor("state")
class StateMachine(SystemWideDevice, ModeDevice):

    """A generic state machine."""

    config_section = 'state_machines'
    collection = 'state_machines'
    class_label = 'state_machine'

    __slots__ = ["player", "_state", "_handlers", "_show"]

    def __init__(self, machine, name):
        """Initialise state machine."""
        super().__init__(machine, name)
        self.player = None
        self._state = None
        self._handlers = []
        self._show = None

    async def device_added_system_wide(self):
        """Initialise internal state."""
        await super().device_added_system_wide()

        if self.config['persist_state']:
            self.raise_config_error("Cannot set persist_state for system-wide state_machine", 1)

        self._start_state(self.config['starting_state'])

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None):
        """Validate transitions."""
        result = super().validate_and_parse_config(config, is_mode_config, debug_prefix)
        states = config.get("states", {})
        for transition in config.get("transitions", []):
            for state in transition["source"]:
                if state not in states:
                    self.raise_config_error("Source {} of transition {} not found in states.".format(
                        transition["source"], transition), 2)
            if transition["target"] not in states:
                self.raise_config_error("Target {} of transition {} not found in states.".format(
                    transition["target"], transition), 3)
        return result

    @property
    def can_exist_outside_of_game(self) -> bool:
        """Return true if persist_state is not set."""
        return not self.config['persist_state']

    @property
    def state(self):
        """Return the current state."""
        if self.config['persist_state'] and self.player:
            return self.player["state_machine_{}".format(self.name)]

        return self._state

    @state.setter
    def state(self, value):
        """Set the current state."""
        old = self.state
        if self.config['persist_state']:
            old = self.player["state_machine_{}".format(self.name)]
            self.player["state_machine_{}".format(self.name)] = value
            self.notify_virtual_change(self, old, value)
        else:
            self._state = value

        # notify monitors
        self.notify_virtual_change("state", old, self.state)

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Restore internal state from player if persist_state is set or create new state."""
        super().device_loaded_in_mode(mode, player)
        self.player = player
        if not self.state:
            self._start_state(self.config['starting_state'])
        else:
            self._add_handlers_for_current_state()
            self._run_show_for_current_state()

    def device_removed_from_mode(self, mode: Mode):
        """Unset internal state to prevent leakage."""
        super().device_removed_from_mode(mode)
        self._remove_handlers()
        self._state = None
        self.player = None

        if self._show:
            self._show.stop()
            self._show = None

    def _stop_current_state(self):
        self.log.debug("Stopping state %s", self.state)
        self._remove_handlers()
        state_config = self.config['states'][self.state]
        if state_config['events_when_stopped']:
            for event_name in state_config['events_when_stopped']:
                self.machine.events.post(event_name)

        if self._show:
            self.log.debug("Stopping show %s", self._show)
            self._show.stop()
            self._show = None

        self.state = None

    def _start_state(self, state):
        self.log.debug("Starting state %s", state)
        if state not in self.config['states']:
            raise AssertionError("Invalid state {}".format(state))

        state_config = self.config['states'][state]
        self.state = state
        if state_config['events_when_started']:
            for event_name in state_config['events_when_started']:
                self.machine.events.post(event_name)

        self._add_handlers_for_current_state()
        self._run_show_for_current_state()

    def _run_show_for_current_state(self):
        assert not self._show
        state_config = self.config['states'][self.state]
        if state_config['show_when_active']:
            self.log.debug("Starting show %s", state_config['show_when_active'])
            self._show = self.machine.show_controller.play_show_with_config(state_config['show_when_active'],
                                                                            self.mode)

    def _add_handlers_for_current_state(self):
        for transition in self.config['transitions']:
            if self.state in transition['source']:
                for event in transition['events']:
                    self._handlers.append(self.machine.events.add_handler(event, self._transition,
                                                                          transition_config=transition))

    def _transition(self, transition_config, **kwargs):
        del kwargs
        self.log.info("Transitioning from %s to %s", self.state, transition_config["target"])
        self._stop_current_state()
        if transition_config['events_when_transitioning']:
            for event_name in transition_config['events_when_transitioning']:
                self.machine.events.post(event_name)
        self._start_state(transition_config['target'])

    def _remove_handlers(self):
        self.machine.events.remove_handlers_by_keys(self._handlers)



"""Device that implements an extra ball group."""
from typing import Optional

from mpf.core.events import event_handler
from mpf.core.machine import MachineController
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.system_wide_device import SystemWideDevice
from mpf.core.player import Player


@DeviceMonitor("enabled")
class ExtraBallGroup(SystemWideDevice):

    """Tracks and manages groups of extra balls devices."""

    config_section = 'extra_ball_groups'
    collection = 'extra_ball_groups'
    class_label = 'extra_ball_group'

    __slots__ = ["player"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialize ExtraBallGroup."""
        super().__init__(machine, name)

        self.player = None  # type: Optional[Player]
        self.machine.events.add_handler('player_added',
                                        self._player_added)
        self.machine.events.add_handler('player_turn_starting',
                                        self._player_turn_starting)
        self.machine.events.add_handler('player_turn_ending',
                                        self._player_turn_ending)
        self.machine.events.add_handler('ball_started',
                                        self._ball_started)

    @property
    def enabled(self):
        """Return whether this extra ball group is enabled.

        This attribute considers the enabled setting plus the
        max balls per game and ball settings.
        """
        if not self.player or not self.config['enabled']:
            return False

        if (self.config['max_per_game'] and self.config['max_per_game'] <=
                self.player['extra_ball_group_{}_num_awarded_game'.format(
                    self.name)]):
            return False

        if (self.config['max_per_ball'] and self.config['max_per_ball'] <=
                self.player['extra_ball_group_{}_num_awarded_ball'.format(
                    self.name)]):
            return False

        return True

    def _player_added(self, player, **kwargs):
        # called once per player to setup their vars for this group
        del kwargs

        player['extra_ball_group_{}_num_awarded_game'.format(self.name)] = 0
        player['extra_ball_group_{}_num_lit'.format(self.name)] = 0
        player['extra_ball_group_{}_num_awarded_ball'.format(self.name)] = 0

    def _player_turn_starting(self, player, number, **kwargs):
        # reset the num of EBs awarded per ball. We do this on turn start
        # rather than ball start because a player shooting again is technically
        # another ball start even though it's the same ball number
        del number, kwargs

        self.player = player
        player['extra_ball_group_{}_num_awarded_ball'.format(self.name)] = 0

    def _ball_started(self, ball, player, **kwargs):
        # check if we need to relight the group
        del ball, player, kwargs
        if self.player['extra_ball_group_{}_num_lit'.format(self.name)]:
            self._post_lit_events()

    def _player_turn_ending(self, player, number, **kwargs):
        # clear the lit if lit memory is disabled
        del number, kwargs

        if not self.config['lit_memory']:
            player['extra_ball_group_{}_num_lit'.format(self.name)] = 0

        self.player = None

    def is_ok_to_light(self) -> bool:
        """Check if it's possible to light an extra ball.

        Returns True or False.

        This method checks to see if the group is enabled and whether the
        max_lit setting has been exceeded.
        """
        if not self.enabled or not self.player:
            return False

        return not (
            self.config['max_lit'] and self.config['max_lit'] <=
            self.player['extra_ball_group_{}_num_lit'.format(self.name)])

    @event_handler(2)
    def event_award_lit(self, **kwargs):
        """Handle award_lit control event."""
        del kwargs
        self.award_lit()

    def award_lit(self):
        """Award a lit extra ball.

        If the player does not have any lit extra balls, this method does
        nothing.
        """
        if not self.player:
            return

        if not self.enabled:
            self.award_disabled()
            return

        if self.player['extra_ball_group_{}_num_lit'.format(self.name)] < 1:
            return

        self.player['extra_ball_group_{}_num_lit'.format(self.name)] -= 1

        if not self.player['extra_ball_group_{}_num_lit'.format(self.name)]:
            self._post_unlit_events()
            posted_unlit_events = True
        else:
            posted_unlit_events = True

        self.award(posted_unlit_events=posted_unlit_events)

    @event_handler(1)
    def event_award(self, posted_unlit_events=False, **kwargs):
        """Handle award control event."""
        del kwargs
        self.award(posted_unlit_events)

    def award(self, posted_unlit_events=False):
        """Immediately awards an extra ball.

        This event first checks to make sure the limits of the max extra
        balls have not been exceeded and that this group is enabled.

        Note that this method will work even if this group does not have any
        extra balls or extra balls lit. You can use this to directly award an
        extra ball.
        """
        if not self.enabled:
            self.award_disabled()
            return

        self.machine.events.post('extra_ball_group_{}_awarded'.format(self.name))
        '''event: extra_ball_group_(name)_awarded

        desc: An extra ball from this group was just awarded. This is a
        good event to use to trigger award shows, sounds, etc.
        '''
        self.player['extra_ball_group_{}_num_awarded_game'.format(self.name)] += 1
        self.player['extra_ball_group_{}_num_awarded_ball'.format(self.name)] += 1
        self.player.extra_balls += 1
        self.machine.events.post('extra_ball_awarded')

        # if this award puts us over the max limits, make sure none are lit
        if not self.enabled:
            self.player['extra_ball_group_{}_num_lit'] = 0
            if not posted_unlit_events:
                self._post_unlit_events()

    @event_handler(3)
    def event_light(self, **kwargs):
        """Handle light control event."""
        del kwargs
        self.light()

    def light(self):
        """Light the extra ball for possible collection by the player.

        This method checks that the group is enabled and that the max lit
        value has not been exceeded. If so, this method will post the extra
        ball disabled events.
        """
        if self.is_ok_to_light():
            self.player['extra_ball_group_{}_num_lit'.format(self.name)] += 1

            self.machine.events.post(
                'extra_ball_group_{}_lit_awarded'.format(self.name))
            '''event: extra_ball_group_(name)_lit_awarded

            desc: This even is posted when an extra ball is lit during play.
            It is NOT posted when a player's turn starts if they have a lit
            extra ball from their previous turn. Therefore this event is a
            good event to use for your award slides and shows when a player
            lights the extra ball, because you don't want to use
            :doc:`extra_ball_group_extra_ball_group_lit` because that is also posted when
            the player's turn starts and you don't want the award show to play
            again when they're starting their turn.
            '''

            self._post_lit_events()

        else:
            self._extra_ball_award_disabled()

    def _post_lit_events(self, **kwargs):
        del kwargs

        self.machine.events.post(
            'extra_ball_group_{}_lit'.format(self.name))
        '''event: extra_ball_group_(name)_lit

        desc: An extra ball was just lit. This is a good event to use to
        start your extra ball lit mode, to turn on an extra ball light,
        to play the "get that extra ball" sound, etc.

        Note that this event is posted if an extra ball is lit during play
        and also when a player's turn starts if they have a lit extra ball.

        See also the :doc:`extra_ball_extra_ball_lit` for a similar event that
        is only posted when an extra ball is lit during play, and not
        if the player starts their turn with the extra ball lit.
        '''

    def _post_unlit_events(self, **kwargs):
        del kwargs

        self.machine.events.post(
            'extra_ball_group_{}_unlit'.format(self.name))

        '''event: extra_ball_group_(name)_unlit

        desc: No more lit extra balls are available for this extra ball group.
        This is a good event to
        use as a stop event for your extra ball lit mode or whatever you're
        using to indicate to the player that an extra ball is available.
        '''

    def award_disabled(self):
        """Post the events when an extra ball connect be awarded."""
        self.machine.events.post('extra_ball_group_{}_award_disabled'.format(self.name))
        '''event: extra_ball_group_(name)_award_disabled

        desc: Posted when you have the global extra ball settings set to not
        enable extra balls but where an extra ball would have been awarded.
        This is a good alternative event to use to score points or whatever
        else you want to give the player when extra balls are disabled.

        '''



from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase

from mpf.tests.MpfGameTestCase import MpfGameTestCase
from unittest.mock import MagicMock


class TestGame(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/game/'

    def get_platform(self):
        return 'smart_virtual'

    def testSinglePlayerGame(self):
        # setup event callbacks
        self._events = MagicMock()

        # Create handler entries for all game lifecycle events we wish to test
        self.machine.events.add_handler('game_will_start', self._events, event_name='game_will_start')
        self.machine.events.add_handler('game_starting', self._events, event_name='game_starting')
        self.machine.events.add_handler('game_started', self._events, event_name='game_started')
        self.machine.events.add_handler('player_add_request', self._events, event_name='player_add_request')
        self.machine.events.add_handler('player_will_add', self._events, event_name='player_will_add')
        self.machine.events.add_handler('player_adding', self._events, event_name='player_adding')
        self.machine.events.add_handler('player_added', self._events, event_name='player_added')
        self.machine.events.add_handler('player_turn_will_start', self._events, event_name='player_turn_will_start')
        self.machine.events.add_handler('player_turn_starting', self._events, event_name='player_turn_starting')
        self.machine.events.add_handler('player_turn_started', self._events, event_name='player_turn_started')
        self.machine.events.add_handler('ball_will_start', self._events, event_name='ball_will_start')
        self.machine.events.add_handler('ball_starting', self._events, event_name='ball_starting')
        self.machine.events.add_handler('ball_started', self._events, event_name='ball_started')
        self.machine.events.add_handler('ball_will_end', self._events, event_name='ball_will_end')
        self.machine.events.add_handler('ball_ending', self._events, event_name='ball_ending')
        self.machine.events.add_handler('ball_ended', self._events, event_name='ball_ended')
        self.machine.events.add_handler('game_will_end', self._events, event_name='game_will_end')
        self.machine.events.add_handler('game_ending', self._events, event_name='game_ending')
        self.machine.events.add_handler('game_ended', self._events, event_name='game_ended')

        # prepare game
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        # start game (single player)
        self.start_game()
        self.assertGameIsRunning()
        self.assertPlayerNumber(1)
        self.assertBallNumber(1)

        self.assertEqual(3, self.machine.modes["game"].balls_per_game)

        # Assert game startup sequence
        self.assertEqual(13, self._events.call_count)
        self.assertEqual('game_will_start', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('game_starting', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('player_add_request', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_will_add', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_adding', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_added', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[5][1]['num'])
        self.assertEqual('game_started', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('player_turn_will_start', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual('player_turn_starting', self._events.call_args_list[8][1]['event_name'])
        self.assertEqual('player_turn_started', self._events.call_args_list[9][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[9][1]['number'])
        self.assertEqual('ball_will_start', self._events.call_args_list[10][1]['event_name'])
        self.assertEqual('ball_starting', self._events.call_args_list[11][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[11][1]['balls_remaining'])
        self.assertFalse(self._events.call_args_list[11][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[12][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[12][1]['ball'])
        self.assertEqual(1, self._events.call_args_list[12][1]['player'])
        self._events.reset_mock()

        # Drain the first ball
        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertPlayerNumber(1)
        self.assertBallNumber(2)

        # Assert ball drain, next ball start sequence
        self.assertEqual(9, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_turn_will_start', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_turn_starting', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_turn_started', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[5][1]['number'])
        self.assertEqual('ball_will_start', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('ball_starting', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[7][1]['balls_remaining'])
        self.assertFalse(self._events.call_args_list[7][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[8][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[8][1]['ball'])
        self._events.reset_mock()

        # Drain the second ball
        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertPlayerNumber(1)
        self.assertBallNumber(3)

        # Assert ball drain, next ball start sequence
        self.assertEqual(9, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_turn_will_start', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_turn_starting', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_turn_started', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[5][1]['number'])
        self.assertEqual('ball_will_start', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('ball_starting', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual(0, self._events.call_args_list[7][1]['balls_remaining'])
        self.assertFalse(self._events.call_args_list[7][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[8][1]['event_name'])
        self.assertEqual(3, self._events.call_args_list[8][1]['ball'])
        self._events.reset_mock()

        # Drain the third (and last) ball
        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertGameIsNotRunning()

        # Assert ball drain, game ending sequence
        self.assertEqual(6, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('game_will_end', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('game_ending', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('game_ended', self._events.call_args_list[5][1]['event_name'])

    def testMultiplePlayerGame(self):
        # setup event callbacks
        self._events = MagicMock()

        # Create handler entries for all game lifecycle events we wish to test
        self.machine.events.add_handler('game_will_start', self._events, event_name='game_will_start')
        self.machine.events.add_handler('game_starting', self._events, event_name='game_starting')
        self.machine.events.add_handler('game_started', self._events, event_name='game_started')
        self.machine.events.add_handler('player_add_request', self._events, event_name='player_add_request')
        self.machine.events.add_handler('player_will_add', self._events, event_name='player_will_add')
        self.machine.events.add_handler('player_adding', self._events, event_name='player_adding')
        self.machine.events.add_handler('player_added', self._events, event_name='player_added')
        self.machine.events.add_handler('player_turn_will_start', self._events, event_name='player_turn_will_start')
        self.machine.events.add_handler('player_turn_starting', self._events, event_name='player_turn_starting')
        self.machine.events.add_handler('player_turn_started', self._events, event_name='player_turn_started')
        self.machine.events.add_handler('player_turn_will_end', self._events, event_name='player_turn_will_end')
        self.machine.events.add_handler('player_turn_ending', self._events, event_name='player_turn_ending')
        self.machine.events.add_handler('player_turn_ended', self._events, event_name='player_turn_ended')
        self.machine.events.add_handler('ball_will_start', self._events, event_name='ball_will_start')
        self.machine.events.add_handler('ball_starting', self._events, event_name='ball_starting')
        self.machine.events.add_handler('ball_started', self._events, event_name='ball_started')
        self.machine.events.add_handler('ball_will_end', self._events, event_name='ball_will_end')
        self.machine.events.add_handler('ball_ending', self._events, event_name='ball_ending')
        self.machine.events.add_handler('ball_ended', self._events, event_name='ball_ended')
        self.machine.events.add_handler('game_will_end', self._events, event_name='game_will_end')
        self.machine.events.add_handler('game_ending', self._events, event_name='game_ending')
        self.machine.events.add_handler('game_ended', self._events, event_name='game_ended')

        # prepare game
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        # start game (first player)
        self.start_game()
        self.advance_time_and_run(5)
        self.assertGameIsRunning()
        self.assertPlayerNumber(1)
        self.assertBallNumber(1)

        self.assertEqual(3, self.machine.modes["game"].balls_per_game)

        # Assert game startup sequence
        self.assertEqual(13, self._events.call_count)
        self.assertEqual('game_will_start', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('game_starting', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('player_add_request', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_will_add', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_adding', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_added', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[5][1]['num'])
        self.assertEqual('game_started', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('player_turn_will_start', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual('player_turn_starting', self._events.call_args_list[8][1]['event_name'])
        self.assertEqual('player_turn_started', self._events.call_args_list[9][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[9][1]['number'])
        self.assertEqual('ball_will_start', self._events.call_args_list[10][1]['event_name'])
        self.assertEqual('ball_starting', self._events.call_args_list[11][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[11][1]['balls_remaining'])
        self.assertFalse(self._events.call_args_list[11][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[12][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[12][1]['ball'])
        self.assertEqual(1, self._events.call_args_list[12][1]['player'])
        self._events.reset_mock()

        # add another player (player 2)
        self.add_player()

        # Assert game startup sequence
        self.assertEqual(4, self._events.call_count)
        self.assertEqual('player_add_request', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('player_will_add', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('player_adding', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_added', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[3][1]['num'])
        self._events.reset_mock()

        # Drain the first ball (player 1)
        self.drain_all_balls()
        self.advance_time_and_run(5)
        self.assertPlayerNumber(2)
        self.assertBallNumber(1)

        # Assert ball drain, next ball start sequence
        self.assertEqual(12, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_turn_will_end', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_turn_ending', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_turn_ended', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual('player_turn_will_start', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('player_turn_starting', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual('player_turn_started', self._events.call_args_list[8][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[8][1]['number'])
        self.assertEqual('ball_will_start', self._events.call_args_list[9][1]['event_name'])
        self.assertEqual('ball_starting', self._events.call_args_list[10][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[10][1]['balls_remaining'])
        self.assertFalse(self._events.call_args_list[10][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[11][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[11][1]['ball'])
        self._events.reset_mock()

        # Drain the first ball (player 2)
        self.drain_all_balls()
        self.advance_time_and_run(5)
        self.assertPlayerNumber(1)
        self.assertBallNumber(2)

        # Assert ball drain, next ball start sequence
        self.assertEqual(12, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_turn_will_end', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_turn_ending', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_turn_ended', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual('player_turn_will_start', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('player_turn_starting', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual('player_turn_started', self._events.call_args_list[8][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[8][1]['number'])
        self.assertEqual('ball_will_start', self._events.call_args_list[9][1]['event_name'])
        self.assertEqual('ball_starting', self._events.call_args_list[10][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[10][1]['balls_remaining'])
        self.assertFalse(self._events.call_args_list[10][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[11][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[11][1]['ball'])
        self._events.reset_mock()

        # Drain the second ball (player 1)
        self.drain_all_balls()
        self.advance_time_and_run(5)
        self.assertPlayerNumber(2)
        self.assertBallNumber(2)

        # Assert ball drain, next ball start sequence
        self.assertEqual(12, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_turn_will_end', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_turn_ending', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_turn_ended', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual('player_turn_will_start', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('player_turn_starting', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual('player_turn_started', self._events.call_args_list[8][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[8][1]['number'])
        self.assertEqual('ball_will_start', self._events.call_args_list[9][1]['event_name'])
        self.assertEqual('ball_starting', self._events.call_args_list[10][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[10][1]['balls_remaining'])
        self.assertFalse(self._events.call_args_list[10][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[11][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[11][1]['ball'])
        self._events.reset_mock()

        # Player 2 earns extra ball before draining
        self.machine.game.player.extra_balls += 1

        # Drain the ball (player 2 has earned an extra ball so it should still be
        # player 2's turn)
        self.drain_all_balls()
        self.advance_time_and_run(5)
        self.assertPlayerNumber(2)
        self.assertBallNumber(2)

        # Assert ball drain, next ball sequence
        self.assertEqual(6, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('ball_will_start', self._events.call_args_list[3][1]['event_name'])
        self.assertTrue(self._events.call_args_list[3][1]['is_extra_ball'])
        self.assertEqual('ball_starting', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[4][1]['balls_remaining'])
        self.assertTrue(self._events.call_args_list[4][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[5][1]['ball'])
        self._events.reset_mock()

        # Drain the second ball (player 2)
        self.drain_all_balls()
        self.advance_time_and_run(5)
        self.assertPlayerNumber(1)
        self.assertBallNumber(3)

        # Assert ball drain, next ball start sequence
        self.assertEqual(12, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_turn_will_end', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_turn_ending', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_turn_ended', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual('player_turn_will_start', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('player_turn_starting', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual('player_turn_started', self._events.call_args_list[8][1]['event_name'])
        self.assertEqual(1, self._events.call_args_list[8][1]['number'])
        self.assertEqual('ball_will_start', self._events.call_args_list[9][1]['event_name'])
        self.assertEqual('ball_starting', self._events.call_args_list[10][1]['event_name'])
        self.assertEqual(0, self._events.call_args_list[10][1]['balls_remaining'])
        self.assertFalse(self._events.call_args_list[10][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[11][1]['event_name'])
        self.assertEqual(3, self._events.call_args_list[11][1]['ball'])
        self._events.reset_mock()

        # Drain the third ball (player 1)
        self.drain_all_balls()
        self.advance_time_and_run(5)
        self.assertPlayerNumber(2)
        self.assertBallNumber(3)

        # Assert ball drain, next ball start sequence
        self.assertEqual(12, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_turn_will_end', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_turn_ending', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_turn_ended', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual('player_turn_will_start', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('player_turn_starting', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual('player_turn_started', self._events.call_args_list[8][1]['event_name'])
        self.assertEqual(2, self._events.call_args_list[8][1]['number'])
        self.assertEqual('ball_will_start', self._events.call_args_list[9][1]['event_name'])
        self.assertEqual('ball_starting', self._events.call_args_list[10][1]['event_name'])
        self.assertEqual(0, self._events.call_args_list[10][1]['balls_remaining'])
        self.assertFalse(self._events.call_args_list[10][1]['is_extra_ball'])
        self.assertEqual('ball_started', self._events.call_args_list[11][1]['event_name'])
        self.assertEqual(3, self._events.call_args_list[11][1]['ball'])
        self._events.reset_mock()

        # Drain the third (and last) ball for player 2
        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertGameIsNotRunning()

        # Assert ball drain, game ending sequence
        self.assertEqual(9, self._events.call_count)
        self.assertEqual('ball_will_end', self._events.call_args_list[0][1]['event_name'])
        self.assertEqual('ball_ending', self._events.call_args_list[1][1]['event_name'])
        self.assertEqual('ball_ended', self._events.call_args_list[2][1]['event_name'])
        self.assertEqual('player_turn_will_end', self._events.call_args_list[3][1]['event_name'])
        self.assertEqual('player_turn_ending', self._events.call_args_list[4][1]['event_name'])
        self.assertEqual('player_turn_ended', self._events.call_args_list[5][1]['event_name'])
        self.assertEqual('game_will_end', self._events.call_args_list[6][1]['event_name'])
        self.assertEqual('game_ending', self._events.call_args_list[7][1]['event_name'])
        self.assertEqual('game_ended', self._events.call_args_list[8][1]['event_name'])

    def testGameEvents(self):
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        self.post_event("start_my_game")
        self.assertGameIsRunning()
        self.advance_time_and_run()
        self.assertPlayerCount(1)
        self.post_event("start_my_game")
        self.assertPlayerCount(1)

        self.post_event("add_my_player")
        self.assertPlayerCount(2)
        self.post_event("add_my_player")
        self.assertPlayerCount(3)
        self.post_event("add_my_player")
        self.assertPlayerCount(4)
        self.post_event("add_my_player")
        self.assertPlayerCount(4)


class TestGameLogic(MpfFakeGameTestCase):

    def testLastGameScore(self):
        # no previous scores
        self.assertFalse(self.machine.variables.is_machine_var("player1_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player2_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player3_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player4_score"))

        # four players
        self.start_game()
        self.add_player()
        self.add_player()
        self.add_player()
        self.machine.game.player.score = 100
        self.assertPlayerNumber(1)
        self.drain_all_balls()
        self.machine.game.player.score = 200
        self.assertPlayerNumber(2)
        self.drain_all_balls()
        self.machine.game.player.score = 0
        self.assertPlayerNumber(3)
        self.drain_all_balls()
        self.machine.game.player.score = 42
        self.assertPlayerNumber(4)

        # still old scores should not be set
        self.assertFalse(self.machine.variables.is_machine_var("player1_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player2_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player3_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player4_score"))

        self.stop_game()

        self.assertMachineVarEqual(100, "player1_score")
        self.assertMachineVarEqual(200, "player2_score")
        self.assertMachineVarEqual(0, "player3_score")
        self.assertMachineVarEqual(42, "player4_score")

        # two players
        self.start_game()
        self.add_player()
        self.machine.game.player.score = 100
        self.assertPlayerNumber(1)
        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.machine.game.player.score = 200
        self.drain_all_balls()
        # old scores should still be active
        self.assertMachineVarEqual(100, "player1_score")
        self.assertMachineVarEqual(200, "player2_score")
        self.assertMachineVarEqual(0, "player3_score")
        self.assertMachineVarEqual(42, "player4_score")
        self.stop_game()

        self.assertMachineVarEqual(100, "player1_score")
        self.assertMachineVarEqual(200, "player2_score")
        self.assertFalse(self.machine.variables.is_machine_var("player3_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player4_score"))

        # start one player game
        self.start_game()
        self.machine.game.player.score = 1337
        self.drain_all_balls()
        self.drain_all_balls()
        # still the old scores
        self.assertMachineVarEqual(100, "player1_score")
        self.assertMachineVarEqual(200, "player2_score")
        self.assertFalse(self.machine.variables.is_machine_var("player3_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player4_score"))
        self.drain_all_balls()
        self.assertGameIsNotRunning()

        self.assertMachineVarEqual(1337, "player1_score")
        self.assertFalse(self.machine.variables.is_machine_var("player2_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player3_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player4_score"))



"""A platform to control lights via OSC."""
try:
    from pythonosc.dispatcher import Dispatcher
    from pythonosc.osc_server import AsyncIOOSCUDPServer
    from pythonosc.udp_client import SimpleUDPClient
except ImportError:
    # pythonosc is not a requirement for MPF so we fail with a nice error when loading
    Dispatcher = None
    AsyncIOOSCUDPServer = None
    SimpleUDPClient = None

from typing import Dict

from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface
from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface, LightPlatformSoftwareFade
from mpf.core.platform import LightsPlatform, SwitchPlatform, SwitchConfig


class OscLight(LightPlatformSoftwareFade):

    """A light on OSC."""

    def __init__(self, number, loop, software_fade_ms, client):
        """Initialise light."""
        super().__init__(number, loop, software_fade_ms)
        self.client = client    # type: SimpleUDPClient

    def get_board_name(self):
        """Return board name."""
        return "OSC light"

    def set_brightness(self, brightness: float):
        """Set brightness of OSC light."""
        self.client.send_message("/light/" + self.number, brightness)

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        raise AssertionError("Not possible in OSC")

    def get_successor_number(self):
        """Return next number."""
        raise AssertionError("Not possible in OSC")

    def __lt__(self, other):
        """Order lights by string."""
        return self.number < other.number


class OscSwitch(SwitchPlatformInterface):

    """A switch in the OSC platform."""

    def get_board_name(self):
        """Return board name."""
        return "OSC Switch"


class OscPlatform(LightsPlatform, SwitchPlatform):

    """A platform to control lights via OSC."""

    __slots__ = ["config", "client", "switches", "dispatcher", "server"]

    def __init__(self, machine):
        """Initialize OSC platform."""
        super().__init__(machine)
        self.configure_logging("OSC")
        self.config = None
        self.client = None      # type: SimpleUDPClient
        self.server = None
        self.switches = {}      # type: Dict[str, OscSwitch]
        if not SimpleUDPClient:
            raise AssertionError("python-osc is not installed. Please run 'pip3 install python-osc'.")

    # noinspection PyCallingNonCallable
    async def initialize(self):
        """Initialise platform."""
        self.config = self.machine.config['osc']
        self.machine.config_validator.validate_config("osc", self.config)
        self.client = SimpleUDPClient(self.config['remote_ip'], self.config['remote_port'])

        dispatcher = Dispatcher()
        dispatcher.map("/sw/*", self._handle_switch)
        dispatcher.map("/event/*", self._handle_event)
        server = AsyncIOOSCUDPServer((self.config['listen_ip'], self.config['listen_port']), dispatcher,
                                     self.machine.clock.loop)
        self.server, _ = await server.create_serve_endpoint()

        for event in self.config['events_to_send']:
            self.machine.events.add_handler(event, self._send_event, _event_name=event)

    def stop(self):
        """Stop server."""
        self.server.close()

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse light number to three RGB channels."""
        del subtype
        return [
            {
                "number": number + "/red"
            },
            {
                "number": number + "/green"
            },
            {
                "number": number + "/blue"
            },
        ]

    def configure_light(self, number: str, subtype: str, config, platform_settings: dict) -> LightPlatformInterface:
        """Configure an OSC light."""
        del config
        return OscLight(number, self.machine.clock.loop,
                        int(1 / self.machine.config['mpf']['default_light_hw_update_hz'] * 1000), self.client)

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> "SwitchPlatformInterface":
        """Config an OSC switch."""
        switch = OscSwitch(config, str(number))
        self.switches[str(number)] = switch
        return switch

    async def get_hw_switch_states(self) -> Dict[str, bool]:
        """Return all switches as false."""
        result = {}
        for switch in self.switches.values():
            result[switch.number] = False
        return result

    def _handle_switch(self, address, *args):
        """Handle switch change from OSC."""
        try:
            _, _, switch_name = address.split("/")
        except ValueError:
            self.error_log("Got OSC switch change for invalid address: %s", address)
            return

        if len(args) != 1:
            self.error_log("Got OSC switch change with too many arguments: %s", args)
            return

        if not isinstance(args[0], (bool, int)):
            self.error_log("Got OSC switch change with invalid state: %s", args[0])
            return

        if switch_name not in self.switches:
            self.error_log("Got OSC switch change for switch %s which is not configured as OSC switch.", switch_name)
            return

        self.machine.switch_controller.process_switch_by_num(switch_name, bool(args[0]), self, logical=True)

    def _send_event(self, _event_name, **kwargs):
        """Send event to OSC client."""
        params = []
        for key, value in sorted(kwargs.items(), key=lambda t: t[0]):
            params.append(key)
            # OSC supports int, float and string only
            if isinstance(value, (int, str, float)):
                params.append(value)
            else:
                params.append(str(value))
        self.client.send_message("/event/{}".format(_event_name), params)

    def _handle_event(self, address, *args):
        """Handle event from OSC."""
        try:
            _, _, event_name = address.split("/")
        except ValueError:
            self.error_log("Got OSC event for invalid address: %s", address)
            return

        if len(args) % 2 != 0:
            self.error_log("Got OSC event with an uneven number of arguments: %s. Arguments need to be pairs "
                           "key1, value1, key2, value2 and so on.", args)
            return

        kwargs = {}
        for i in range(0, len(args), 2):
            kwargs[args[i]] = args[i + 1]

        self.machine.events.post(event_name, **kwargs)



"""Interface for config file loaders."""
import logging
import os
from typing import Union, Tuple, Optional

MYPY = False
if MYPY:    # pragma: no cover
    from typing import List     # pylint: disable-msg=cyclic-import,unused-import


class FileInterface:

    """Interface for config files."""

    __slots__ = ["log"]

    file_types = list()     # type: List[str]

    def __init__(self):
        """Initialise file manager."""
        self.log = logging.getLogger('{} File Interface'.format(
            self.file_types[0][1:].upper()))

    def find_file(self, filename) -> Tuple[Optional[str], Optional[str]]:
        """Test whether the passed file is valid.

        If the file does not have an externsion, this method will test for files with that base name with
        all the extensions it can read.

        Args:
        ----
            filename: Full absolute path of a file to check, with or without
                an extension.

        Returns false if a file is not found. Tuple of (full file with path, extension) if a file is found.
        """
        if not os.path.splitext(filename)[1]:
            # file has no extension

            for extension in self.file_types:
                if os.path.isfile(filename + extension):
                    return os.path.abspath(filename + extension), extension
            return None, None

        return filename, os.path.splitext(filename)[1]

    def load(self, filename, expected_version_str=None, halt_on_error=True):
        """Load file."""
        raise NotImplementedError

    def save(self, filename, data):
        """Save file."""
        raise NotImplementedError
